@@ -3,7 +3,7 @@
  *
  * This file is part of fizmo.
  *
- * Copyright (c) 2009-2012 Christoph Ender.
+ * Copyright (c) 2009-2013 Christoph Ender.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -45,6 +45,7 @@
 #include <signal.h>
 #include <ncurses.h>
 //#include <ncursesw/ncurses.h>
+#include <wchar.h>
 
 #include <tools/i18n.h>
 #include <tools/tracelog.h>
@@ -68,7 +69,7 @@
 
 #include "../locales/fizmo_ncursesw_locales.h"
 
-#define FIZMO_NCURSESW_VERSION "0.7.2"
+#define FIZMO_NCURSESW_VERSION "0.7.7"
 
 #ifdef ENABLE_X11_IMAGES
 #include <drilbo/drilbo.h>
@@ -315,7 +316,6 @@ static void infowin_z_ucs_output_wordwrap_destination(z_ucs *z_ucs_output,
 }
 
 
-#ifdef ENABLE_TRACING
 static z_colour curses_to_z_colour(short curses_color)
 {
   switch (curses_color)
@@ -332,7 +332,6 @@ static z_colour curses_to_z_colour(short curses_color)
 
   return -1;
 }
-#endif // ENABLE_TRACING
 
 
 static void goto_yx(int y, int x)
@@ -643,6 +642,12 @@ static void print_startup_syntax()
       i18n_ncursesw_DISABLE_HYPHENATION);
   streams_latin1_output("\n");
 
+  streams_latin1_output( " -mu, --maximum-undo-steps: ");
+  i18n_translate(
+      fizmo_ncursesw_module_name,
+      i18n_ncursesw_SET_NUMBER_OF_MAXIMUM_UNDO_STEPS);
+  streams_latin1_output("\n");
+
 #ifdef ENABLE_X11_IMAGES
   streams_latin1_output( " -nx, --disable-x11-graphics: ");
   i18n_translate(
@@ -1568,12 +1573,14 @@ static int get_next_event(z_ucs *z_ucs_input, int timeout_millis)
             result = EVENT_WAS_CODE_PAGE_DOWN;
           else if (input == KEY_PPAGE)
             result = EVENT_WAS_CODE_PAGE_UP;
+          else if (input == KEY_DC)
+            result = EVENT_WAS_CODE_DELETE;
           else if (input == KEY_BACKSPACE)
             result = EVENT_WAS_CODE_BACKSPACE;
         }
         else if (input_return_code == OK)
         {
-          if ( (input == 127) || (input == 8) )
+          if (input == 127)
             result = EVENT_WAS_CODE_BACKSPACE;
           else if (input == 1)
             result = EVENT_WAS_CODE_CTRL_A;
@@ -1841,13 +1848,20 @@ static void set_cursor_visibility(bool visible)
 
 static z_colour get_default_foreground_colour()
 {
-  return Z_COLOUR_WHITE;
+  return screen_default_foreground_color;
 }
 
 
 static z_colour get_default_background_colour()
 {
-  return Z_COLOUR_BLACK;
+  return screen_default_background_color;
+}
+
+static int prompt_for_filename(char *UNUSED(filename_suggestion),
+    z_file **UNUSED(result_file), char *UNUSED(directory),
+    int UNUSED(filetype_or_mode), int UNUSED(fileaccess))
+{
+  return -3;
 }
 
 
@@ -1882,7 +1896,8 @@ static struct z_screen_cell_interface ncursesw_interface =
   &clear_area,
   &set_cursor_visibility,
   &get_default_foreground_colour,
-  &get_default_background_colour
+  &get_default_background_colour,
+  &prompt_for_filename
 };
 
 
@@ -2270,6 +2285,9 @@ static char *select_story_from_menu()
   endwin();
   wordwrap_destroy_wrapper(infowin_output_wordwrapper);
 
+  free(infowin_more);
+  free(infowin_back);
+
   return result;
 }
 
@@ -2324,11 +2342,15 @@ int main(int argc, char *argv[])
   char *cwd = NULL;
   char *absdirname = NULL;
   z_colour new_color;
+  chtype default_bkgd;
 #ifndef DISABLE_FILELIST
   char *story_to_load_filename, *assumed_filename;
   struct z_story_list *story_list;
   size_t absdirname_len = 0;
-  int i;
+  int i, default_pair;
+  SCREEN *s;
+  short pair_foreground, pair_background;
+  z_colour default_fore, default_back;
 #endif // DISABLE_FILELIST
 
 #ifdef ENABLE_TRACING
@@ -2338,6 +2360,38 @@ int main(int argc, char *argv[])
   setlocale(LC_ALL, "C");
   setlocale(LC_CTYPE, "");
 
+  // newterm instead of initscr to init ncurses without clearing screen
+  // so we can exit gracefully in case of errors.
+  if ((s = newterm(NULL, stdin, stdout)) == NULL)
+    return -1;
+  if (has_colors() == true) {
+    start_color();
+    default_bkgd = getbkgd(stdscr);
+    default_pair = default_bkgd & A_COLOR;
+    TRACE_LOG("Default pair: %d\n", default_pair);
+    if (pair_content(default_pair, &pair_foreground, &pair_background) == ERR) {
+      TRACE_LOG("pair_content returned ERR.\n");
+      screen_default_foreground_color = Z_COLOUR_WHITE;
+      screen_default_background_color = Z_COLOUR_BLACK;
+    }
+    else {
+      default_fore = curses_to_z_colour(pair_foreground);
+      default_back = curses_to_z_colour(pair_background);
+      if ((default_fore < 0) || (default_back < 0)) {
+        TRACE_LOG("Illegal default color values, reverting to white/black.\n");
+        screen_default_foreground_color = Z_COLOUR_WHITE;
+        screen_default_background_color = Z_COLOUR_BLACK;
+      }
+      else {
+        TRACE_LOG("Default colors from ncurses: %d, %d.\n",
+            default_fore, default_back);
+        screen_default_foreground_color = default_fore;
+        screen_default_background_color = default_back;
+      }
+    }
+  }
+  endwin();
+
   fizmo_register_screen_cell_interface(&ncursesw_interface);
 
 #ifdef SOUND_INTERFACE_STRUCT_NAME
@@ -2462,13 +2516,18 @@ int main(int argc, char *argv[])
         exit(EXIT_FAILURE);
       }
 
-      if ((new_color = colorname_to_infocomcode(argv[argi])) == -1)
-      {
+      if (set_configuration_value("background-color", (argv[argi])) < 0) {
         print_startup_syntax();
         exit(EXIT_FAILURE);
       }
 
+      if ((new_color = colorname_to_infocomcode(argv[argi])) == -1) {
+        // Shouldn't happen since set-config said the value is okay.
+        exit(EXIT_FAILURE);
+      }
+
       screen_default_background_color = new_color;
+
       argi++;
     }
     else if (
@@ -2481,13 +2540,18 @@ int main(int argc, char *argv[])
         exit(EXIT_FAILURE);
       }
 
-      if ((new_color = colorname_to_infocomcode(argv[argi])) == -1)
-      {
+      if (set_configuration_value("foreground-color", (argv[argi])) < 0) {
         print_startup_syntax();
         exit(EXIT_FAILURE);
       }
 
+      if ((new_color = colorname_to_infocomcode(argv[argi])) == -1) {
+        // Shouldn't happen since set-config said the value is okay.
+        exit(EXIT_FAILURE);
+      }
+
       screen_default_foreground_color = new_color;
+
       argi++;
     }
     else if (
@@ -2580,11 +2644,15 @@ int main(int argc, char *argv[])
     else if (
         (strcmp(argv[argi], "-lm") == 0)
         ||
-        (strcmp(argv[argi], "-rm") == 0)
-        ||
         (strcmp(argv[argi], "--left-margin") == 0)
         ||
+        (strcmp(argv[argi], "-rm") == 0)
+        ||
         (strcmp(argv[argi], "--right-margin") == 0)
+        ||
+        (strcmp(argv[argi], "-mu") == 0)
+        ||
+        (strcmp(argv[argi], "--maximum-undo-steps") == 0)
         )
     {
       if (++argi == argc)
@@ -2619,8 +2687,23 @@ int main(int argc, char *argv[])
           (strcmp(argv[argi - 1], "--left-margin") == 0)
          )
         set_custom_left_cell_margin(int_value);
-      else
+      else if (
+          (strcmp(argv[argi - 1], "-rm") == 0)
+          ||
+          (strcmp(argv[argi - 1], "--right-margin") == 0)
+          )
         set_custom_right_cell_margin(int_value);
+      else if (
+          (strcmp(argv[argi - 1], "-mu") == 0)
+          ||
+          (strcmp(argv[argi - 1], "--maximum-undo-steps") == 0)
+          )
+        set_configuration_value("max-undo-steps", argv[argi]);
+      else
+      {
+        // Internal mismatch, must not happen since we've already verified
+        // these values above.
+      }
 
       argi += 1;
     }
@@ -2648,7 +2731,11 @@ int main(int argc, char *argv[])
         (strcmp(argv[argi], "--update-story-list") == 0))
     {
       printf("\n");
-      update_fizmo_story_list();
+      story_list = update_fizmo_story_list();
+      // Only update functionality is relevant, so we can forget the
+      // result right away.
+      if (story_list != NULL)
+        free_z_story_list(story_list);
       printf("\n");
       directory_was_searched = true;
       argi += 1;
@@ -2881,9 +2968,7 @@ int main(int argc, char *argv[])
   fizmo_start(
       story_stream,
       blorb_stream,
-      savegame_to_restore,
-      screen_default_foreground_color,
-      screen_default_background_color);
+      savegame_to_restore);
 
   sigaction(SIGWINCH, NULL, NULL);
 
@@ -2895,6 +2980,10 @@ int main(int argc, char *argv[])
   close(ncursesw_if_signalling_pipe[1]);
   close(ncursesw_if_signalling_pipe[0]);
 
+#ifdef ENABLE_TRACING
+  turn_off_trace();
+#endif // ENABLE_TRACING
+
   return 0;
 }
 
