@@ -108,7 +108,7 @@ static int active_z_window_id = -1;
 static z_colour current_output_foreground_colour = -3;
 static z_colour current_output_background_colour = -3;
 static z_style current_output_text_style = -1;
-static WORDWRAP *refresh_wordwrapper;
+static WORDWRAP *refresh_wordwrapper = NULL;
 static int refresh_newline_counter;
 static bool refresh_count_mode;
 static int refresh_lines_to_skip;
@@ -1119,6 +1119,7 @@ static int cell_close_interface(z_ucs *error_message)
 {
   int event_type;
   z_ucs input;
+  int i;
 
   if ( (error_message == NULL) && (interface_open == true) )
   {
@@ -1136,7 +1137,39 @@ static int cell_close_interface(z_ucs *error_message)
   }
 
   screen_cell_interface->close_interface(error_message);
+
+  free(libcellif_more_prompt);  
+  free(libcellif_score_string);
+  free(libcellif_turns_string);
+
+  if (refresh_wordwrapper != NULL)
+  {
+    wordwrap_destroy_wrapper(refresh_wordwrapper);
+    refresh_wordwrapper = NULL;
+  }
+
+  for (i=0; i<nof_active_z_windows; i++)
+  {
+    if (z_windows[i]->wordwrapper != NULL)
+    {
+      if (z_windows[i] != NULL)
+      {
+        wordwrap_destroy_wrapper(z_windows[i]->wordwrapper);
+        z_windows[i]->wordwrapper = NULL;
+        free(z_windows[i]);
+        z_windows[i] = NULL;
+      }
+    }
+  }
+
+  if (z_windows != NULL)
+  {
+    free(z_windows);
+    z_windows = NULL;
+  }
+
   interface_open = false;
+
   return 0;
 }
 
@@ -1622,7 +1655,7 @@ static void refresh_screen()
       block_index = i * upper_window_buffer->width;
 
       j = 0;
-      screen_cell_interface->goto_yx(1+i, 1);
+      screen_cell_interface->goto_yx(i + (ver <= 3 ? 2 : 1), 1);
 
       while (
           (j < z_windows[1]->xsize)
@@ -1722,6 +1755,8 @@ static void refresh_screen()
         }
       }
     }
+
+    free(blockbuf_line);
   }
 
   update_output_colours(0);
@@ -1974,7 +2009,7 @@ static int16_t read_line(zscii *dest, uint16_t maximum_length,
   int input_display_width; // Width of the input line on-screen.
   int input_x, input_y; // Leftmost position of the input line on-screen.
   z_ucs input_buffer[maximum_length + 1];
-  int cmd_history_index = 0, cmd_index;
+  int cmd_history_index = 0;
   zscii *cmd_history_ptr;
   int current_tenth_seconds = 0;
   int timed_routine_retval, index;
@@ -2015,8 +2050,6 @@ static int16_t read_line(zscii *dest, uint16_t maximum_length,
 
   TRACE_LOG("Flush finished.\n");
 
-  timeout_millis = (is_timed_keyboard_input_available() == true ? 100 : 0);
-
   TRACE_LOG("1/10s: %d, routine: %d.\n",
       tenth_seconds, verification_routine);
 
@@ -2033,7 +2066,10 @@ static int16_t read_line(zscii *dest, uint16_t maximum_length,
       *tenth_seconds_elapsed = 0;
   }
   else
+  {
+    timeout_millis = 0;
     timed_input_active = false;
+  }
 
   screen_cell_interface->update_screen();
   update_output_colours(active_z_window_id);
@@ -2154,6 +2190,13 @@ static int16_t read_line(zscii *dest, uint16_t maximum_length,
           //refresh_screen();
           screen_cell_interface->redraw_screen_from_scratch();
         }
+        else if (input == 18)
+        {
+          TRACE_LOG("Got CTRL-R.\n");
+          new_cell_screen_size(
+              screen_cell_interface->get_screen_height(),
+              screen_cell_interface->get_screen_width());
+        }
         else if (
             // Check if we have a valid input char.
             (unicode_char_to_zscii_input_char(input) != 0xff)
@@ -2294,58 +2337,53 @@ static int16_t read_line(zscii *dest, uint16_t maximum_length,
                 ((z_windows[active_z_window_id]->xpos
                   + z_windows[active_z_window_id]->xcursorpos - 1) - input_x));
 
-            /*
-            // Check if there's now enough input to fill the whole line.
-            if (input_size - input_scroll_x < input_display_width)
-            {
-            // If there's not enough input, we'll have to fill the column
-            // directly right after the last char with a space.
-            char_buf[0] = Z_UCS_SPACE;
-            }
-            else
-            {
-            // In case there's enough input we'll take the char to fill the
-            // rightmost input column to fill with from the input.
-            char_buf[0] = 'X';
-            }
-            screen_cell_interface->goto_yx(input_y,
-            input_x + input_display_width - 1);
-            screen_cell_interface->z_ucs_output(char_buf);
-            screen_cell_interface->goto_yx(input_y,
-            input_x + input_display_width - 1);
-            */
-
             z_windows[active_z_window_id]->xcursorpos--;
             refresh_cursor(active_z_window_id);
 
             screen_cell_interface->update_screen();
           }
+        }
+      }
+      else if (event_type == EVENT_WAS_CODE_DELETE)
+      {
+        // We only have something to do if the cursor is not after the end of
+        // input.
+        if (input_index < input_size)
+        {
+          // We always have to move all chars from the cursor position onward
+          // one position to the left.
+          memmove(
+              input_buffer + input_index,
+              input_buffer + input_index + 1,
+              sizeof(z_ucs)*(input_size - input_index));
 
+          input_size--;
 
-          /*
-             if (z_windows[active_z_window_id]->xpos
-             + z_windows[active_z_window_id]->xcursorpos -1
-             > input_x)
-             z_windows[active_z_window_id]->xcursorpos--;
+          TRACE_LOG("Moving %d chars from  %d/%d to %d/%d.\n",
+              input_display_width -
+              ((z_windows[active_z_window_id]->xpos
+                + z_windows[active_z_window_id]->xcursorpos) - input_x),
+              z_windows[active_z_window_id]->xpos
+              + z_windows[active_z_window_id]->xcursorpos,
+              input_y,
+              z_windows[active_z_window_id]->xpos
+              + z_windows[active_z_window_id]->xcursorpos - 1,
+              input_y);
 
-          // The position of the newly created space (from the copy_area above)
-          // on the screen is stored in pos.
-          pos
-          = z_windows[active_z_window_id]->xpos
-          + z_windows[active_z_window_id]->xcursorpos - 1
-          + (input_size - input_index);
-          maxpos
-          = z_windows[active_z_window_id]->xpos
-          + z_windows[active_z_window_id]->xsize - 1
-          - z_windows[active_z_window_id]->rightmargin;
-          if (pos > maxpos)
-          pos = maxpos;
+          screen_cell_interface->copy_area(
+              input_y,
+              z_windows[active_z_window_id]->xpos
+              + z_windows[active_z_window_id]->xcursorpos - 1,
+              input_y,
+              z_windows[active_z_window_id]->xpos
+              + z_windows[active_z_window_id]->xcursorpos,
+              1,
+              input_display_width -
+              ((z_windows[active_z_window_id]->xpos
+                + z_windows[active_z_window_id]->xcursorpos) - input_x));
 
-          screen_cell_interface->goto_yx(input_y, pos);
-          char_buf[0] = 'X';
-          screen_cell_interface->z_ucs_output(char_buf);
           refresh_cursor(active_z_window_id);
-          */
+          screen_cell_interface->update_screen();
         }
       }
       else if (event_type == EVENT_WAS_CODE_CURSOR_LEFT)
@@ -2453,7 +2491,7 @@ static int16_t read_line(zscii *dest, uint16_t maximum_length,
            (
             (event_type == EVENT_WAS_CODE_CURSOR_UP)
             &&
-            (cmd_history_index < command_history_nof_entries)
+            (cmd_history_index < get_number_of_stored_commands())
            )
            ||
            (
@@ -2467,18 +2505,8 @@ static int16_t read_line(zscii *dest, uint16_t maximum_length,
         TRACE_LOG("old history index: %d.\n", cmd_history_index);
 
         cmd_history_index += event_type == EVENT_WAS_CODE_CURSOR_UP ? 1 : -1;
-        if (cmd_history_index - 1 > command_history_newest_entry)
-          cmd_index
-            =  command_history_nof_entries
-            - (cmd_history_index - 1)
-            -  command_history_newest_entry;
-        else
-          cmd_index
-            = command_history_newest_entry - (cmd_history_index - 1);
-
-        cmd_history_ptr
-          = command_history_buffer
-          + command_history_entries[cmd_index];
+        cmd_history_ptr = get_command_from_history(cmd_history_index - 1);
+        TRACE_LOG("cmd_history_ptr: %p.\n", cmd_history_ptr);
 
         if (cmd_history_index > 0)
         {
@@ -2721,6 +2749,16 @@ static int read_char(uint16_t tenth_seconds, uint32_t verification_routine,
         result = 132;
         input_in_progress = false;
       }
+      else if (event_type == EVENT_WAS_CODE_BACKSPACE)
+      {
+        result = 8;
+        input_in_progress = false;
+      }
+      else if (event_type == EVENT_WAS_CODE_DELETE)
+      {
+        result = 127;
+        input_in_progress = false;
+      }
       else if (event_type == EVENT_WAS_TIMEOUT)
       {
         TRACE_LOG("timeout found.\n");
@@ -3007,6 +3045,15 @@ static void game_was_restored_and_history_modified()
 }
 
 
+static int prompt_for_filename(char *filename_suggestion,
+    z_file **result_file, char *directory, int filetype_or_mode,
+    int fileaccess)
+{
+  return screen_cell_interface->prompt_for_filename(filename_suggestion,
+      result_file, directory, filetype_or_mode, fileaccess);
+}
+
+
 static struct z_screen_interface z_cell_interface =
 {
   &get_interface_name,
@@ -3055,7 +3102,10 @@ static struct z_screen_interface z_cell_interface =
   &erase_line_pixels,
   &output_interface_info,
   &input_must_be_repeated_by_story,
-  &game_was_restored_and_history_modified
+  &game_was_restored_and_history_modified,
+  &prompt_for_filename,
+  NULL, /* do_autosave */
+  NULL, /* restore_autosave */
 };
 
 
@@ -3172,6 +3222,7 @@ void new_cell_screen_size(int newysize, int newxsize)
       z_windows[i]->ycursorpos = z_windows[i]->ysize;
 
     TRACE_LOG("new ycursorpos[%d]: %d\n", i, z_windows[i]->ycursorpos);
+    TRACE_LOG("new ypos[%d]: %d\n", i, z_windows[i]->ypos);
 
     if (z_windows[i]->xcursorpos > z_windows[i]->xsize)
       z_windows[i]->xcursorpos = z_windows[i]->xsize;
@@ -3188,35 +3239,16 @@ void new_cell_screen_size(int newysize, int newxsize)
     *current_input_display_width
       = z_windows[0]->xpos + z_windows[0]->xsize - *current_input_x;
 
-    /*
-    *current_input_y += dy;
-
-    if (*current_input_y > z_windows[0]->ypos + z_windows[0]->ysize - 1)
-      *current_input_y = z_windows[0]->ypos + z_windows[0]->ysize - 1;
-    */
-
     // If the screen is redrawn, the screen contents are always aligned
     // to the bottom so we also have to move the input line downward.
     *current_input_y = z_windows[0]->ypos + z_windows[0]->ysize - 1;
 
-    /*
-       current_input_size = &input_size;
-       current_input_scroll_x = &input_scroll_x;
-       current_input_index = &input_index;
-       current_input_display_width = &input_display_width;
-       current_input_x = &input_x;
-       current_input_y = &input_y;
-       current_input_buffer = input_buffer;
-     */
-
     TRACE_LOG("current_input_display_width: %d\n",*current_input_display_width);
 
     TRACE_LOG("input-x-after: %d\n", *current_input_x);
     TRACE_LOG("input-y-after: %d\n", *current_input_y);
   }
 
-
-  //wordwrap_output_left_padding(z_windows[i]->wordwrapper);
   refresh_screen();
 
   for (i=0; i<nof_active_z_windows; i++)
