@@ -3,7 +3,7 @@
  *
  * This file is part of fizmo.
  *
- * Copyright (c) 2009-2012 Christoph Ender.
+ * Copyright (c) 2009-2013 Christoph Ender.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -165,7 +165,6 @@ static int add_doc_to_babel_info(xmlDocPtr new_babel_doc,
       // Neither "/ifindex" nor "/if:ifindex" found. Skip this file.
       xmlXPathFreeObject(xpathObj);
       xmlXPathFreeContext(xpathCtx); 
-      xmlFreeDoc(new_babel_doc);
       return -1;
     }
     else
@@ -209,7 +208,7 @@ struct babel_info *load_babel_info_from_blorb(z_file *infile, int length,
   char *xmlData = (char*)fizmo_malloc(length + 1);
   xmlDocPtr babel_doc;
 
-  if (fsi->getchars(xmlData, length, infile) != (size_t)length)
+  if (fsi->readchars(xmlData, length, infile) != (size_t)length)
   {
     free(xmlData);
     return NULL;
@@ -217,16 +216,16 @@ struct babel_info *load_babel_info_from_blorb(z_file *infile, int length,
 
   xmlData[length] = '\0';
 
-  if ((babel_doc = xmlReadDoc(
-          (xmlChar*)xmlData,
-          NULL,
-          NULL,
-          XML_PARSE_NOWARNING | XML_PARSE_NOERROR))
-      == NULL)
-  {
-    free(xmlData);
+  babel_doc = xmlReadDoc(
+      (xmlChar*)xmlData,
+      NULL,
+      NULL,
+      XML_PARSE_NOWARNING | XML_PARSE_NOERROR);
+
+  free(xmlData);
+
+  if (babel_doc == NULL)
     return NULL;
-  }
 
   result = (struct babel_info*)fizmo_malloc(sizeof(struct babel_info));
   result->entries = NULL;
@@ -236,8 +235,8 @@ struct babel_info *load_babel_info_from_blorb(z_file *infile, int length,
   if (add_doc_to_babel_info(babel_doc, result, last_mod_timestamp, filename)
       != 0)
   {
+    xmlFreeDoc(babel_doc);
     free(result);
-    free(xmlData);
     return NULL;
   }
 
@@ -324,6 +323,7 @@ struct babel_info *load_babel_info()
                 new_babel_doc, result, last_mod_timestamp, z_dir_entry.d_name))
             != 0)
         {
+          xmlFreeDoc(new_babel_doc);
           free_babel_info(result);
           fsi->ch_dir(cwd);
           free(cwd);
@@ -578,7 +578,10 @@ void store_babel_info_timestamps(struct babel_info *babel)
 
   if ((out = fsi->openfile(filename, FILETYPE_DATA, FILEACCESS_WRITE))
       == NULL)
+  {
+    free(filename);
     return;
+  }
 
   if (babel)
   {
@@ -591,6 +594,7 @@ void store_babel_info_timestamps(struct babel_info *babel)
     }
   }
 
+  free(filename);
   fsi->closefile(out);
 }
 
@@ -634,17 +638,21 @@ bool babel_files_have_changed(struct babel_info *babel)
 
   if ((timestamp_file = fsi->openfile(
           filename, FILETYPE_DATA, FILEACCESS_READ)) == NULL)
-    return true;
-
-  if ((data = fsi->getchar(timestamp_file)) != EOF)
   {
-    fsi->ungetchar(data, timestamp_file);
+    free(filename);
+    return true;
+  }
+
+  if ((data = fsi->readchar(timestamp_file)) != EOF)
+  {
+    fsi->unreadchar(data, timestamp_file);
     for(;;)
     {
       offset = fsi->getfilepos(timestamp_file);
-      while ((data = fsi->getchar(timestamp_file)) != '\t')
+      while ((data = fsi->readchar(timestamp_file)) != '\t')
         if (data == EOF)
         {
+          free(filename);
           abort_timestamp_input();
           return true;
         }
@@ -653,6 +661,7 @@ bool babel_files_have_changed(struct babel_info *babel)
       if (ensure_mem_size(&timestamp_input, &timestamp_input_size, size + 2)
           == -1)
       {
+        free(filename);
         abort_timestamp_input();
         return true;
       }
@@ -660,9 +669,10 @@ bool babel_files_have_changed(struct babel_info *babel)
       if (size > 0)
       {
         fsi->setfilepos(timestamp_file, -(size+1), SEEK_CUR);
-        if (fsi->getchars(timestamp_input, size+1, timestamp_file)
+        if (fsi->readchars(timestamp_input, size+1, timestamp_file)
             != (size_t)size+1)
         {
+          free(filename);
           abort_timestamp_input();
           return true;
         }
@@ -670,16 +680,18 @@ bool babel_files_have_changed(struct babel_info *babel)
       timestamp_input[size] = '\0';
 
       offset = fsi->getfilepos(timestamp_file);
-      while ((data = fsi->getchar(timestamp_file)) != '\n')
+      while ((data = fsi->readchar(timestamp_file)) != '\n')
         if (data == EOF)
         {
+          free(filename);
           abort_timestamp_input();
           return true;
         }
 
       size = fsi->getfilepos(timestamp_file) - offset - 1;
-      if (ensure_mem_size(&filename_input, &filename_input_size, size + 2) == -1)
+      if (ensure_mem_size(&filename_input, &filename_input_size, size+2) == -1)
       {
+        free(filename);
         abort_timestamp_input();
         return true;
       }
@@ -687,9 +699,10 @@ bool babel_files_have_changed(struct babel_info *babel)
       if (size > 0)
       {
         fsi->setfilepos(timestamp_file, -(size+1), SEEK_CUR);
-        if (fsi->getchars(filename_input, size+1, timestamp_file)
+        if (fsi->readchars(filename_input, size+1, timestamp_file)
             != (size_t)size+1)
         {
+          free(filename);
           abort_timestamp_input();
           return true;
         }
@@ -710,6 +723,7 @@ bool babel_files_have_changed(struct babel_info *babel)
             if (isdigit(timestamp_input[j]) == 0)
             {
               abort_timestamp_input();
+              free(filename);
               return true;
             }
 
@@ -722,6 +736,7 @@ bool babel_files_have_changed(struct babel_info *babel)
           {
             //printf("Timestamps dont match\n");
             abort_timestamp_input();
+            free(filename);
             return true;
           }
 
@@ -733,17 +748,19 @@ bool babel_files_have_changed(struct babel_info *babel)
       {
         abort_timestamp_input();
         //printf("File not found: %s\n", filename_input);
+        free(filename);
         return true;
       }
 
       nof_babel_timestamp_entries++;
 
-      if ((data = fsi->getchar(timestamp_file)) == EOF)
+      if ((data = fsi->readchar(timestamp_file)) == EOF)
         break;
-      fsi->ungetchar(data, timestamp_file);
+      fsi->unreadchar(data, timestamp_file);
     }
   }
 
+  free(filename);
   abort_timestamp_input();
 
   if (nof_babel_timestamp_entries != (babel ? babel->nof_entries : 0))
