@@ -45,7 +45,7 @@
 #include "../tools/tracelog.h"
 #include "config.h"
 #include "fizmo.h"
-#include "math.h"
+#include "mathemat.h"
 #include "output.h"
 #include "../tools/types.h"
 #include "../tools/i18n.h"
@@ -67,27 +67,32 @@ char *user_homedir = NULL;
 struct configuration_option configuration_options[] = {
 
   // String values:
+  { "autosave-filename", NULL },
   { "background-color", NULL },
-  { "input-command-filename", NULL },
-  { "record-command-filename", NULL },
   { "foreground-color", NULL },
   { "i18n-search-path", NULL },
+  { "input-command-filename", NULL },
   { "locale", NULL },
+  { "max-undo-steps", NULL },
   { "random-mode", NULL },
+  { "record-command-filename", NULL },
   { "save-text-history-paragraphs", NULL },
-  { "savegame-path", NULL },
   { "savegame-default-filename", NULL },
+  { "savegame-path", NULL },
+  { "stream-2-left-margin", NULL },
+  { "stream-2-line-width", NULL },
   { "transcript-filename", NULL },
   { "z-code-path", NULL },
   { "z-code-root-path", NULL },
-  { "stream-2-line-width", NULL },
-  { "stream-2-left-margin", NULL },
 
   // Boolean values:
   { "disable-external-streams", NULL },
   { "disable-restore", NULL },
   { "disable-save", NULL },
   { "disable-sound", NULL },
+  { "disable-stream-2-hyphenation", NULL },
+  { "disable-stream-2-wrap", NULL },
+  { "dont-set-locale-from-config", NULL },
   { "enable-font3-conversion", NULL },
   { "quetzal-umem", NULL },
   { "restore-after-save-and-quit-file-before-read", NULL },
@@ -96,9 +101,7 @@ struct configuration_option configuration_options[] = {
   { "start-command-recording-when-story-starts", NULL },
   { "start-file-input-when-story-starts", NULL },
   { "start-script-when-story-starts", NULL },
-  { "disable-stream-2-hyphenation", NULL },
   { "sync-transcript", NULL },
-  { "dont-set-locale-from-config", NULL },
 
   // NULL terminates the option list.
   { NULL, NULL }
@@ -182,7 +185,10 @@ static char *expand_configuration_value(char *unexpanded_value)
         while ( (*ptr != 0) && (*ptr != ')') )
           ptr++;
         if (*ptr != ')')
+        {
+          free(result);
           return NULL;
+        }
         buf = *ptr;
         *ptr = 0;
 
@@ -321,6 +327,8 @@ int set_configuration_value(char *key, char* new_unexpanded_value)
           ||
           (strcmp(key, "z-code-root-path") == 0)
           ||
+          (strcmp(key, "autosave-filename") == 0)
+          ||
           (strcmp(key, "savegame-path") == 0)
           ||
           (strcmp(key, "savegame-default-filename") == 0)
@@ -345,16 +353,25 @@ int set_configuration_value(char *key, char* new_unexpanded_value)
           (strcmp(key, "stream-2-line-width") == 0)
           ||
           (strcmp(key, "stream-2-left-margin") == 0)
+          ||
+          (strcmp(key, "max-undo-steps") == 0)
           )
       {
-        if ( (new_value == NULL) || (strlen(new_value) == 0) )
+        if (new_value == NULL)
           return -1;
+        if (strlen(new_value) == 0)
+        {
+          free(new_value);
+          return -1;
+        }
         strtol(new_value, &endptr, 10);
         if (*endptr != 0)
         {
           free(new_value);
           return -1;
         }
+        if (configuration_options[i].value != NULL)
+          free(configuration_options[i].value);
         configuration_options[i].value = new_value;
         return 0;
       }
@@ -362,28 +379,32 @@ int set_configuration_value(char *key, char* new_unexpanded_value)
       // Color options
       else if (strcmp(key, "foreground-color") == 0)
       {
-        if ((color_code = color_name_to_z_colour(new_value)) == -1)
-        {
-          free(new_value);
+        if (new_value == NULL)
           return -1;
-        }
+        color_code = color_name_to_z_colour(new_value);
         free(new_value);
+        if (color_code == -1)
+          return -1;
         if (snprintf(buf, BUFSIZE, "%d", color_code) >= BUFSIZE)
           return -1;
+        if (configuration_options[i].value != NULL)
+          free(configuration_options[i].value);
         configuration_options[i].value = fizmo_strdup(buf);
         default_foreground_colour = color_code;
         return 0;
       }
       else if (strcmp(key, "background-color") == 0)
       {
-        if ((color_code = color_name_to_z_colour(new_value)) == -1)
-        {
-          free(new_value);
+        if (new_value == NULL)
           return -1;
-        }
+        color_code = color_name_to_z_colour(new_value);
         free(new_value);
+        if (color_code == -1)
+          return -1;
         if (snprintf(buf, BUFSIZE, "%d", color_code) >= BUFSIZE)
           return -1;
+        if (configuration_options[i].value != NULL)
+          free(configuration_options[i].value);
         configuration_options[i].value = fizmo_strdup(buf);
         default_background_colour = color_code;
         return 0;
@@ -419,6 +440,8 @@ int set_configuration_value(char *key, char* new_unexpanded_value)
           ||
           (strcmp(key, "disable-stream-2-hyphenation") == 0)
           ||
+          (strcmp(key, "disable-stream-2-wrap") == 0)
+          ||
           (strcmp(key, "sync-transcript") == 0)
           ||
           (strcmp(key, "dont-set-locale-from-config") == 0)
@@ -571,6 +594,8 @@ char *get_configuration_value(char *key)
             ||
             (strcmp(key, "disable-stream-2-hyphenation") == 0)
             ||
+            (strcmp(key, "disable-stream-2-wrap") == 0)
+            ||
             (strcmp(key, "sync-transcript") == 0)
             ||
             (strcmp(key, "dont-set-locale-from-config") == 0)
@@ -596,6 +621,8 @@ char *get_configuration_value(char *key)
             ||
             (strcmp(key, "z-code-root-path") == 0)
             ||
+            (strcmp(key, "autosave-filename") == 0)
+            ||
             (strcmp(key, "savegame-path") == 0)
             ||
             (strcmp(key, "savegame-default-filename") == 0)
@@ -615,6 +642,8 @@ char *get_configuration_value(char *key)
             (strcmp(key, "stream-2-line-width") == 0)
             ||
             (strcmp(key, "stream-2-left-margin") == 0)
+            ||
+            (strcmp(key, "max-undo-steps") == 0)
             )
         {
           TRACE_LOG("Returning value at %p.\n", configuration_options[i].value);
