@@ -3,7 +3,7 @@
  *
  * This file is part of fizmo.
  *
- * Copyright (c) 2009-2012 Christoph Ender.
+ * Copyright (c) 2009-2013 Christoph Ender.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -44,7 +44,7 @@
 #include "fizmo.h"
 #include "zpu.h"
 #include "text.h"
-#include "math.h"
+#include "mathemat.h"
 #include "stack.h"
 #include "misc.h"
 #include "streams.h"
@@ -55,7 +55,10 @@
 #include "filelist.h"
 #include "routine.h"
 #include "variable.h"
+#include "undo.h"
 #include "blorb.h"
+#include "hyphenation.h"
+#include "undo.h"
 #include "../tools/z_ucs.h"
 #include "../tools/types.h"
 #include "../tools/i18n.h"
@@ -84,14 +87,6 @@ struct z_story *active_z_story;
 struct z_screen_interface *active_interface = NULL;
 struct z_sound_interface *active_sound_interface = NULL;
 
-int ask_user_for_file_default_function(zscii *filename_buffer, int buffer_len,
-    int preload_len, int filetype_or_mode, int fileaccess, z_file **result_file,
-    char *directory);
-int (*ask_user_for_file)(zscii *filename_buffer, int buffer_len,
-    int preload_len, int filetype_or_mode, int fileaccess, z_file **result_file,
-    char *directory
-    ) = &ask_user_for_file_default_function;
-
 uint8_t ver = 0;
 uint8_t *header_extension_table;
 uint8_t header_extension_table_size;
@@ -116,6 +111,7 @@ static bool config_files_were_parsed = false;
 #endif // DISABLE_BLOCKBUFFER
 
 
+// "load_z_story" returns malloc()ed z_story, may be freed using free_z_story().
 static struct z_story *load_z_story(z_file *story_stream, z_file *blorb_stream)
 {
   struct z_story *result;
@@ -236,6 +232,7 @@ static struct z_story *load_z_story(z_file *story_stream, z_file *blorb_stream)
   if (cwd != NULL)
     fsi->ch_dir(cwd);
   free(cwd);
+  free(ptr);
 
   if ((fsi->setfilepos(
           result->z_story_file, result->story_file_exec_offset, SEEK_SET)) != 0)
@@ -246,7 +243,7 @@ static struct z_story *load_z_story(z_file *story_stream, z_file *blorb_stream)
         "setfilepos",
         errno);
 
-  if ((z_file_version = fsi->getchar(result->z_story_file)) == EOF)
+  if ((z_file_version = fsi->readchar(result->z_story_file)) == EOF)
     i18n_translate_and_exit(
         libfizmo_module_name,
         i18n_libfizmo_ERROR_READING_FIRST_STORY_BYTE_FROM_P0S,
@@ -282,7 +279,7 @@ static struct z_story *load_z_story(z_file *story_stream, z_file *blorb_stream)
   TRACE_LOG("Loading %li bytes from \"%s\".\n",
       story_size-1, story_stream->filename);
 
-  if (fsi->getchars(
+  if (fsi->readchars(
         result->memory+1, (size_t)(story_size - 1), result->z_story_file)
       != (size_t)(story_size - 1))
   {
@@ -489,6 +486,20 @@ static struct z_story *load_z_story(z_file *story_stream, z_file *blorb_stream)
 }
 
 
+static void free_z_story(struct z_story *story)
+{
+  free(story->memory);
+  if (story->title != NULL)
+    free(story->title);
+  if (story->blorb_map != NULL)
+    active_blorb_interface->free_blorb_map(story->blorb_map);
+  if (story->blorb_file != NULL)
+    fsi->closefile(story->blorb_file);
+  free(story->absolute_file_name);
+  free(story);
+}
+
+
 /*
 struct z_story_blorb_image *get_image_blorb_index(struct z_story *story,
     int resource_number)
@@ -625,8 +636,6 @@ char *get_xdg_config_dir_name()
     xdg_config_home = fizmo_strdup(config_dir_used);
   }
 
-  // REVISIT: free(xdg_config_home) on end.
-
   xdg_config_dir_name_initialized = true;
 
   return xdg_config_home;
@@ -693,8 +702,24 @@ void ensure_dot_fizmo_dir_exists()
 #endif // DISABLE_CONFIGFILES
 
 
-/*@external@*/ void fizmo_new_screen_size(uint8_t width, uint8_t height)
+void fizmo_new_screen_size(uint8_t width, uint8_t height)
 {
+  if (!z_mem)
+  {
+    /* This shouldn't be called before z_mem is allocated. However, the
+       startup sequence is complicated and I want to be extra careful. */
+    return;
+  }
+
+#ifndef DISABLE_BLOCKBUFFER
+  if ( (ver >= 3) && (upper_window_buffer != NULL)
+      && (upper_window_buffer->height > 0) )
+    blockbuf_resize(
+        upper_window_buffer,
+        (int)width,
+        upper_window_buffer->height);
+#endif // DISABLE_BLOCKBUFFER
+
   if (ver >= 4)
   {
     TRACE_LOG("Writing %d to $20, %d to $21.\n", height, width);
@@ -706,9 +731,6 @@ void ensure_dot_fizmo_dir_exists()
   {
     // 8.4.3: In Version 5 and later, the screen's width and height in
     // units should be written to the words at $22 and $24.
-    // of the header, respectively. In Version 6, these bytes are the
-    // other way round (height in $27, width in $26). The width of a
-    // font is defined as the width of its '0' character.
     TRACE_LOG("Writing %d to $22, %d to $24.\n", width, height);
 
     z_mem[0x22] = width >> 8;
@@ -724,7 +746,7 @@ void write_interpreter_info_into_header()
 {
   uint16_t width, height;
 
-  if (active_interface == NULL)
+  if (active_interface == NULL || z_mem == NULL)
     return;
 
   TRACE_LOG("Linking interface \"%s\" to active story.\n",
@@ -783,8 +805,8 @@ void write_interpreter_info_into_header()
     if (active_interface->is_timed_keyboard_input_available() == true)
       z_mem[1] |= 0x80;
 
-    width = active_interface->get_screen_width();
-    height = active_interface->get_screen_height();
+    width = active_interface->get_screen_width_in_characters();
+    height = active_interface->get_screen_height_in_lines();
 
     fizmo_new_screen_size(width, height);
 
@@ -822,15 +844,36 @@ void write_interpreter_info_into_header()
 }
 
 
-void fizmo_register_screen_interface(struct z_screen_interface
-    *screen_interface)
+int fizmo_register_screen_interface(struct z_screen_interface
+    *screen_interface) //, z_colour screen_default_foreground_color,
+    //z_colour screen_default_background_color)
 {
-  if (active_interface == NULL)
-  {
-    active_interface = screen_interface;
-    register_i18n_stream_output_function(streams_z_ucs_output);
-    register_i18n_abort_function(abort_interpreter);
-  }
+  z_colour default_colour;
+
+  if (active_interface != NULL)
+    return -1;
+
+  default_colour = screen_interface->get_default_foreground_colour();
+  if (is_regular_z_colour(default_colour) == false)
+    return -2;
+  default_foreground_colour = default_colour;
+  if (set_configuration_value("foreground-color",
+        z_colour_names[default_foreground_colour]) != 0)
+    return -3;
+
+  default_colour = screen_interface->get_default_background_colour();
+  if (is_regular_z_colour(default_colour) == false)
+    return -4;
+  default_background_colour = default_colour;
+  if (set_configuration_value("background-color",
+        z_colour_names[default_background_colour]) != 0)
+    return -5;
+
+  active_interface = screen_interface;
+  register_i18n_stream_output_function(streams_z_ucs_output);
+  register_i18n_abort_function(abort_interpreter);
+
+  return 0;
 }
 
 
@@ -850,69 +893,6 @@ void fizmo_register_blorb_interface(
 }
 
 
-void fizmo_register_ask_user_for_file_function(
-    int (*ask_user_for_file_function)(zscii *filename_buffer,
-      int buffer_len, int preload_len, int filetype_or_mode,
-      int fileaccess, z_file **result_file,
-      char *directory))
-{
-  ask_user_for_file = ask_user_for_file_function;
-}
-
-
-int ask_user_for_file_default_function(zscii *filename_buffer, int buffer_len,
-    int preload_len, int filetype_or_mode, int fileaccess, z_file **result_file,
-    char *directory)
-{
-  int input_length;
-  z_ucs filename[buffer_len + 1];
-  char *filename_utf8, *prefixed_filename;
-  int i;
-
-  input_length = active_interface->read_line(
-      (uint8_t*)filename_buffer,
-      buffer_len,
-      0,
-      0,
-      preload_len,
-      NULL,
-      true,
-      true);
-
-  if (input_length == 0)
-    *result_file = NULL;
-
-  if (input_length < 1)
-    return input_length;
-
-  for (i=0; i<(int)input_length; i++)
-    filename[i] = zscii_input_char_to_z_ucs(filename_buffer[i]);
-  filename[i] = 0;
-  filename_utf8 = dup_zucs_string_to_utf8_string(filename);
-
-  if (directory != NULL)
-  {
-    prefixed_filename
-      = fizmo_malloc(strlen(filename_utf8) + strlen(directory) + 2);
-    strcpy(prefixed_filename, directory);
-    strcat(prefixed_filename, "/");
-    strcat(prefixed_filename, filename_utf8);
-  }
-  else
-    prefixed_filename = filename_utf8;
-
-  TRACE_LOG("prefixed filename: \"%s\"\n.", prefixed_filename);
-
-  *result_file = fsi->openfile(prefixed_filename, filetype_or_mode, fileaccess);
-
-  if (directory != NULL)
-    free(prefixed_filename);
-  free(filename_utf8);
-
-  return input_length;
-}
-
-
 #ifndef DISABLE_CONFIGFILES
 static int parse_fizmo_config_file(char *filename)
 {
@@ -927,14 +907,14 @@ static int parse_fizmo_config_file(char *filename)
       == NULL)
     return 0;
 
-  while ((c = fsi->getchar(config_file)) != EOF)
+  while ((c = fsi->readchar(config_file)) != EOF)
   {
     i = 0;
     while ( (c != '=') && (c != '\n') && (c != EOF) )
     {
       if ( (i < MAX_CONFIG_OPTION_LENGTH - 1) && ( ! ((i == 0) && (c == ' '))) )
         key[i++] = (char)c;
-      c = fsi->getchar(config_file);
+      c = fsi->readchar(config_file);
     }
 
     if ( (c == '=') || (c == '\n') )
@@ -947,12 +927,12 @@ static int parse_fizmo_config_file(char *filename)
       if (c == '=')
       {
         do
-          c = fsi->getchar(config_file);
+          c = fsi->readchar(config_file);
         while (c == ' ');
         while ( (c != '\n') && (c != EOF) && (i < MAX_CONFIG_OPTION_LENGTH-1) )
         {
           value[i++] = (char)c;
-          c = fsi->getchar(config_file);
+          c = fsi->readchar(config_file);
         }
 
         while ( (i > 0) && (value[i-1] == ' ') )
@@ -1162,20 +1142,20 @@ char *unquote_special_chars(char *s)
 
 
 void fizmo_start(z_file* story_stream, z_file *blorb_stream,
-    z_file *restore_on_start_file, z_colour screen_default_foreground_color,
-    z_colour screen_default_background_color)
+    z_file *restore_on_start_file)
 {
   char *value;
   bool evaluate_result;
   uint8_t flags2;
+  int val;
   char *str, *default_savegame_filename = DEFAULT_SAVEGAME_FILENAME;
   z_colour default_colour;
 
-
-  //TRACE_LOG("Startup for input filename \"%s\".\n", input_filename);
-
   if (active_interface == NULL)
+  {
+    TRACE_LOG("No active interface.");
     return;
+  }
 
   register_i18n_stream_output_function(
       streams_z_ucs_output);
@@ -1195,14 +1175,6 @@ void fizmo_start(z_file* story_stream, z_file *blorb_stream,
   if ((str = getenv("ZCODE_ROOT_PATH")) != NULL)
     append_path_value("z-code-root-path", str);
 
-  default_colour = active_interface->get_default_foreground_colour();
-  if (is_regular_z_colour(default_colour) == true)
-    default_foreground_colour = default_colour;
-
-  default_colour = active_interface->get_default_background_colour();
-  if (is_regular_z_colour(default_colour) == true)
-    default_background_colour = default_colour;
-
 #ifndef DISABLE_CONFIGFILES
   parse_fizmo_config_files();
 #endif // DISABLE_CONFIGFILES
@@ -1238,6 +1210,11 @@ void fizmo_start(z_file* story_stream, z_file *blorb_stream,
   if ((str = get_configuration_value("savegame-default-filename")) != NULL)
     default_savegame_filename = str;
 
+  val = DEFAULT_MAX_UNDO_STEPS;
+  if ((str = get_configuration_value("max-undo-steps")) != NULL)
+    val = atoi(str);
+  set_max_undo_steps(val);
+
   init_streams(default_savegame_filename);
 
   (void)latin1_string_to_zucs_string(
@@ -1249,32 +1226,6 @@ void fizmo_start(z_file* story_stream, z_file *blorb_stream,
   TRACE_LOG_Z_UCS(last_savegame_filename);
   TRACE_LOG("'.\n");
 
-  if ((str = get_configuration_value("foreground-color")) != NULL)
-  {
-    TRACE_LOG("str-foreground-color: %s\n", str);
-    default_colour = atoi(str);
-    if (is_regular_z_colour(default_colour) == true)
-      default_foreground_colour = default_colour;
-  }
-
-  if ((str = get_configuration_value("background-color")) != NULL)
-  {
-    TRACE_LOG("str-background-color: %s\n", str);
-    default_colour = atoi(str);
-    if (is_regular_z_colour(default_colour) == true)
-      default_background_colour = default_colour;
-  }
-
-  TRACE_LOG("screen_default_foreground_color: %d\n",
-      screen_default_foreground_color);
-  if (is_regular_z_colour(screen_default_foreground_color) == true)
-    default_foreground_colour = screen_default_foreground_color;
-
-  TRACE_LOG("screen_default_background_color: %d\n",
-      screen_default_background_color);
-  if (is_regular_z_colour(screen_default_background_color) == true)
-    default_background_colour = screen_default_background_color;
-
   current_foreground_colour = default_foreground_colour;
   current_background_colour = default_background_colour;
 
@@ -1304,6 +1255,16 @@ void fizmo_start(z_file* story_stream, z_file *blorb_stream,
   active_window_number = 0;
   current_font = Z_FONT_NORMAL;
 
+#ifndef DISABLE_BLOCKBUFFER
+  if (ver >= 3)
+    upper_window_buffer
+      = create_blockbuffer(
+          Z_STYLE_ROMAN,
+          Z_FONT_NORMAL,
+          current_foreground_colour,
+          current_background_colour);
+#endif // DISABLE_BLOCKBUFFER
+
 #ifndef DISABLE_OUTPUT_HISTORY
   outputhistory[0]
     = create_outputhistory(
@@ -1316,11 +1277,23 @@ void fizmo_start(z_file* story_stream, z_file *blorb_stream,
         Z_STYLE_ROMAN);
 #endif /* DISABLE_OUTPUT_HISTORY */
 
+  terminate_interpreter = INTERPRETER_QUIT_NONE;
+
   if ( (ver <= 8) && (ver != 6) )
   {
     while (terminate_interpreter == INTERPRETER_QUIT_NONE)
     {
-      if (restore_on_start_file != NULL)
+      if (restore_on_start_file != NULL && active_interface->restore_autosave) 
+      {
+        /* Use the interface's restore routine. */
+        int res = active_interface->restore_autosave(restore_on_start_file);
+        restore_on_start_file = NULL;
+        if (res)
+          interpret_resume();
+        else
+          interpret_from_address(load_word(z_mem + 0x6));
+      }
+      else if (restore_on_start_file != NULL)
       {
         value = get_configuration_value(
             "restore-after-save-and-quit-file-before-read");
@@ -1380,7 +1353,7 @@ void fizmo_start(z_file* story_stream, z_file *blorb_stream,
         TRACE_LOG("Loading %ld bytes.\n",
             (long int)(active_z_story->high_memory_end + 1 - z_mem));
 
-        if (fsi->getchars(
+        if (fsi->readchars(
               z_mem,
               (size_t)(active_z_story->high_memory_end + 1 - z_mem),
               active_z_story->z_story_file)
@@ -1418,8 +1391,36 @@ void fizmo_start(z_file* story_stream, z_file *blorb_stream,
 
   // Close all streams, this will also close the active interface.
   close_streams(NULL);
+  free_undo_memory();
+  free_hyphenation_memory();
+  free_i18n_memory();
 
-  // TODO: Free memory.
+#ifndef DISABLE_BLOCKBUFFER
+  if (upper_window_buffer != NULL)
+    destroy_blockbuffer(upper_window_buffer);
+  upper_window_buffer = NULL;
+#endif // DISABLE_BLOCKBUFFER
+
+#ifndef DISABLE_OUTPUT_HISTORY
+  destroy_outputhistory(outputhistory[0]);
+#endif // DISABLE_OUTPUT_HISTORY
+
+  free_z_story(active_z_story);
+  active_z_story = NULL;
+  z_mem = NULL;
+
+#ifndef DISABLE_CONFIGFILES
+  if (xdg_config_home != NULL)
+  {
+    free(xdg_config_home);
+    xdg_config_home = NULL;
+  }
+  if (fizmo_config_dir_name != NULL)
+  {
+    free(fizmo_config_dir_name);
+    fizmo_config_dir_name = NULL;
+  }
+#endif // DISABLE_CONFIGFILES
 }
 
 #endif /* fizmo_c_INCLUDED */
