@@ -3,7 +3,7 @@
  *
  * This file is part of fizmo.
  *
- * Copyright (c) 2009-2012 Christoph Ender.
+ * Copyright (c) 2009-2013 Christoph Ender.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -41,8 +41,8 @@
 #include "../blorb_interface/blorb_interface.h"
 #include "blockbuf.h"
 
-#define FIZMO_VERSION "0.7.2"
-#define LIBFIZMO_VERSION "0.7.2"
+#define FIZMO_VERSION "0.7.7"
+#define LIBFIZMO_VERSION "0.7.7"
 
 #define FIZMO_INTERPRETER_NUMBER 6
 /*
@@ -70,24 +70,15 @@ From: http://www.logicalshift.demon.co.uk/unix/zoom/manual/configXwin.html
 #define OBEYS_SPEC_MINOR_REVISION_NUMER 0
 
 
-void fizmo_register_screen_interface(
+int fizmo_register_screen_interface(
     struct z_screen_interface *screen_interface);
 void fizmo_register_sound_interface(
     struct z_sound_interface *sound_interface);
 void fizmo_register_blorb_interface(
     struct z_blorb_interface *blorb_interface);
 
-int (*ask_user_for_file)(zscii *filename_buffer, int buffer_len,
-    int preload_len, int filetype_or_mode, int fileaccess, z_file **result_file,
-    char *directory);
-void fizmo_register_ask_user_for_file_function(
-    int (*ask_user_for_file)(zscii *filename_suggestion,
-      int buffer_len, int preload_len, int filetype_or_mode, int fileaccess,
-      z_file **result_file, char* directory));
-   
 void fizmo_start(z_file* story_stream, z_file *blorb_stream,
-    z_file *restore_on_start_file, z_colour screen_default_foreground_color,
-    z_colour screen_default_background_color);
+    z_file *restore_on_start_file);
 void fizmo_new_screen_size(uint8_t width, uint8_t height);
 
 void write_interpreter_info_into_header();
