@@ -46,12 +46,12 @@
  * or completely filled in case "wrapped_around" is true.
  *
  * Metadata -- font, style and colour attributes -- is directly written
- * into the buffer. In order to distinguish metadata from regular buffer
+ * into this buffer. In order to distinguish metadata from regular buffer
  * contents, metadata is prefixed with a HISTORY_METADATA_ESCAPE character
- * which is 0. Regular buffer contents can never contain a plain 0, since
- * 0 is used as a string-terminator. Thus, when writing output into the
- * buffer, 0 characters don't have to be explicitely escaped since these
- * are never written.
+ * which is 0. Since 0 is used as a string-terminator, regular buffer contents
+ * can never contain a plain 0. Thus, when writing output into the buffer, 0
+ * characters don't have to be explicitely escaped since these are never
+ * written.
  *
  * In order to use the history one can create a *history_output using
  * "init_history_output" which will point to the current end of the history.
@@ -71,6 +71,8 @@
 #include "../tools/tracelog.h"
 #include "../tools/types.h"
 #include "../tools/z_ucs.h"
+#include "../tools/i18n.h"
+#include "../locales/libfizmo_locales.h"
 #include "history.h"
 #include "fizmo.h"
 #include "config.h"
@@ -214,7 +216,11 @@ static void process_buffer_back(OUTPUTHISTORY *h, long nof_zucs_chars)
       else
       {
         TRACE_LOG("Inconsistent history metadata.\n");
-        exit(-1);
+        i18n_translate_and_exit(
+            libfizmo_module_name,
+            i18n_libfizmo_INVALID_PARAMETER_TYPE_P0S,
+            -1,
+            "metadata");
       }
     }
 
@@ -535,7 +541,9 @@ void store_data_in_history(OUTPUTHISTORY *h, z_ucs *data, size_t len,
 void store_z_ucs_output_in_history(OUTPUTHISTORY *h, z_ucs *z_ucs_output)
 {
   size_t len;
+#ifdef ENABLE_TRACING
   z_ucs *cl;
+#endif // ENABLE_TRACING
 
   if (z_ucs_output == NULL)
     return;
@@ -545,6 +553,7 @@ void store_z_ucs_output_in_history(OUTPUTHISTORY *h, z_ucs *z_ucs_output)
 
   store_data_in_history(h, z_ucs_output, len, true);
 
+#ifdef ENABLE_TRACING
   cl = get_current_line(h);
   TRACE_LOG("currentline:|");
   if (cl != NULL)
@@ -553,7 +562,7 @@ void store_z_ucs_output_in_history(OUTPUTHISTORY *h, z_ucs *z_ucs_output)
     free(cl);
   }
   TRACE_LOG("|line end.\n");
-
+#endif // ENABLE_TRACING
 }
 
 
@@ -590,7 +599,11 @@ int store_metadata_in_history(OUTPUTHISTORY *h, int metadata_type, ...)
     // -2 is the lowest allowed value for Z_COLOUR_UNDEFINED, 15 the maximum
     // for all combinations of Z_STYLE.
     TRACE_LOG("Parameter value %d outside valid range.\n", parameter);
-    exit(-1);
+    i18n_translate_and_exit(
+        libfizmo_module_name,
+        i18n_libfizmo_INVALID_PARAMETER_TYPE_P0S,
+        -1,
+        "parameter");
   }
 
   if (metadata_type == HISTORY_METADATA_TYPE_FONT)
@@ -622,7 +635,11 @@ int store_metadata_in_history(OUTPUTHISTORY *h, int metadata_type, ...)
     if ( (parameter < -2) || (parameter > 15 ) )
     {
       TRACE_LOG("Parameter value %d outside valid range.\n", parameter);
-      exit(-1);
+      i18n_translate_and_exit(
+          libfizmo_module_name,
+          i18n_libfizmo_INVALID_PARAMETER_TYPE_P0S,
+          -1,
+          "parameter");
     }
     h->history_buffer_front_index_background = parameter;
     output_buffer[3] = (z_ucs)(parameter + HISTORY_METADATA_DATA_OFFSET);
@@ -1175,7 +1192,11 @@ int output_rewind_paragraph(history_output *output)
       else
       {
         TRACE_LOG("Invalid metadata type %d\n", metadata_type);
-        exit(-1);
+        i18n_translate_and_exit(
+            libfizmo_module_name,
+            i18n_libfizmo_INVALID_PARAMETER_TYPE_P0S,
+            -1,
+            "metadata");
       }
     }
      */
@@ -1300,7 +1321,11 @@ int output_repeat_paragraphs(history_output *output, int n,
         else
         {
           TRACE_LOG("Invalid metadata type %d\n", metadata_type);
-          exit(-1);
+          i18n_translate_and_exit(
+              libfizmo_module_name,
+              i18n_libfizmo_INVALID_PARAMETER_TYPE_P0S,
+              -1,
+              "metadata");
         }
       }
     }
