@@ -3,7 +3,7 @@
  *
  * This file is part of fizmo.
  *
- * Copyright (c) 2009-2012 Christoph Ender.
+ * Copyright (c) 2009-2013 Christoph Ender.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -185,6 +185,59 @@ static int save_stack_frame(uint16_t *current_frame_index,
 }
 
 
+int ask_user_for_file(zscii *filename_buffer, int buffer_len,
+    int preload_len, int filetype_or_mode, int fileaccess, z_file **result_file,
+    char *directory)
+{
+  int input_length;
+  z_ucs filename[buffer_len + 1];
+  char *filename_utf8, *prefixed_filename;
+  int i;
+
+  input_length = active_interface->read_line(
+      (uint8_t*)filename_buffer,
+      buffer_len,
+      0,
+      0,
+      preload_len,
+      NULL,
+      true,
+      true);
+
+  if (input_length == 0)
+    *result_file = NULL;
+
+  if (input_length < 1)
+    return input_length;
+
+  for (i=0; i<(int)input_length; i++)
+    filename[i] = zscii_input_char_to_z_ucs(filename_buffer[i]);
+  filename[i] = 0;
+  filename_utf8 = dup_zucs_string_to_utf8_string(filename);
+
+  if (directory != NULL)
+  {
+    prefixed_filename
+      = fizmo_malloc(strlen(filename_utf8) + strlen(directory) + 2);
+    strcpy(prefixed_filename, directory);
+    strcat(prefixed_filename, "/");
+    strcat(prefixed_filename, filename_utf8);
+  }
+  else
+    prefixed_filename = filename_utf8;
+
+  TRACE_LOG("prefixed filename: \"%s\"\n.", prefixed_filename);
+
+  *result_file = fsi->openfile(prefixed_filename, filetype_or_mode, fileaccess);
+
+  if (directory != NULL)
+    free(prefixed_filename);
+  free(filename_utf8);
+
+  return input_length;
+}
+
+
 static int ask_for_filename(char *filename_suggestion, z_file **result_file,
     char *directory, int filetype_or_mode, int fileaccess)
 {
@@ -193,6 +246,22 @@ static int ask_for_filename(char *filename_suggestion, z_file **result_file,
   bool stream_1_active_buf;
   char *filename_utf8;
   int i;
+  int return_code;
+
+  return_code = active_interface->prompt_for_filename(
+      filename_suggestion,
+      result_file,
+      directory,
+      filetype_or_mode,
+      fileaccess);
+
+  // If return_code is == -3, this function is not implemented in the current
+  // screen interface.
+  if (return_code != -3)
+    return return_code;
+
+  /* There was no prompt_for_file, so the interpreter will have to ask for
+   * a filename directly. */
 
   TRACE_LOG("last:\"");
   TRACE_LOG_Z_UCS(last_savegame_filename);
@@ -263,6 +332,7 @@ static int ask_for_filename(char *filename_suggestion, z_file **result_file,
         "streams_latin1_output");
 
   TRACE_LOG("Removing %d chars from history.\n", length);
+  stream_2_remove_chars(length);
 #ifndef DISABLE_COMMAND_HISTORY
   remove_chars_from_history(outputhistory[active_window_number], length);
 #endif /* DISABLE_COMMAND_HISTORY */
@@ -395,9 +465,10 @@ static int _store_save_or_restore_result(uint16_t result_code)
 }
 
 
-// Returns 0 for failure,  1 for "save succeeded" and 2 for "the game is
-// being restored and is resuming execution again from here, the point where
-// it was saved".
+// Handle an error in the save or restore process. This always returns 0,
+// indicating failure.
+// If evaluate_result is true, this does a Z-machine store of 0.
+// If close_stream is true, this closes the file.
 static int _handle_save_or_restore_failure(bool evaluate_result,
     int i18n_message_code, z_file *iff_file, bool close_stream)
 {
@@ -442,27 +513,14 @@ int get_paragraph_save_amount()
     return strtol(nof_paragraphs_as_string, NULL, 10);
 }
 
-
 void save_game(uint16_t address, uint16_t length, char *filename,
     bool skip_asking_for_filename, bool evaluate_result, char *directory)
 {
-  uint32_t pc_on_restore = (uint32_t)(pc - z_mem);
   z_file *save_file;
-  uint8_t *dynamic_index;
-  uint16_t consecutive_zeros;
-  int data;
   char *system_filename;
-  uint8_t *ptr;
   char *str;
-#ifndef DISABLE_OUTPUT_HISTORY
-  z_ucs *hst_ptr;
-  int nof_paragraphs_to_save;
-  history_output *history;
-  int return_code;
-#endif // DISABLE_OUTPUT_HISTORY
 
   TRACE_LOG("Save %d bytes from address %d.\n", length, address);
-  TRACE_LOG("PC at: %x.\n", pc_on_restore);
 
   if (filename != NULL)
   {
@@ -512,24 +570,44 @@ void save_game(uint16_t address, uint16_t length, char *filename,
     TRACE_LOG("filename from ask_for_filename: \"%s\".\n", str);
   }
 
+  save_game_to_stream(address, length, save_file, evaluate_result);
+}
+
+/* Returns 0 for failure, 1 for success. 
+   This closes the save_file. */
+int save_game_to_stream(uint16_t address, uint16_t length, z_file *save_file,
+  bool evaluate_result)
+{
+  uint32_t pc_on_restore = (uint32_t)(pc - z_mem);
+  uint8_t *dynamic_index;
+  uint16_t consecutive_zeros;
+  int data;
+  uint8_t *ptr;
+#ifndef DISABLE_OUTPUT_HISTORY
+  z_ucs *hst_ptr;
+  int nof_paragraphs_to_save;
+  history_output *history;
+  int return_code;
+#endif // DISABLE_OUTPUT_HISTORY
+
+  TRACE_LOG("PC at: %x.\n", pc_on_restore);
+
   if (address != 0)
   {
     if ((fsi->writechars(z_mem + address, length, save_file)) != length)
     {
-      _handle_save_or_restore_failure(
+      return _handle_save_or_restore_failure(
           evaluate_result,
           i18n_libfizmo_ERROR_WRITING_SAVE_FILE,
           save_file, true);
-      return;
     }
 
     if ((fsi->closefile(save_file)) != 0)
     {
-      _handle_save_or_restore_failure(
+      return _handle_save_or_restore_failure(
           evaluate_result,
           i18n_libfizmo_ERROR_WRITING_SAVE_FILE,
           save_file, true);
-      return;
     }
   }
   else
@@ -538,102 +616,87 @@ void save_game(uint16_t address, uint16_t length, char *filename,
 
     if (start_new_chunk("IFhd", save_file) != 0)
     {
-      _handle_save_or_restore_failure(evaluate_result,
+      return _handle_save_or_restore_failure(evaluate_result,
           i18n_libfizmo_ERROR_WRITING_SAVE_FILE, save_file, true);
-      return;
     }
 
     // Save release number
     if ((fsi->writechar((int)z_mem[0x2], save_file)) == EOF)
     {
-      _handle_save_or_restore_failure(evaluate_result,
+      return _handle_save_or_restore_failure(evaluate_result,
           i18n_libfizmo_ERROR_WRITING_SAVE_FILE, save_file, true);
-      return;
     }
     if ((fsi->writechar((int)z_mem[0x3], save_file)) == EOF)
     {
-      _handle_save_or_restore_failure(evaluate_result,
+      return _handle_save_or_restore_failure(evaluate_result,
           i18n_libfizmo_ERROR_WRITING_SAVE_FILE, save_file, true);
-      return;
     }
 
     // Save serial number
     if ((fsi->writechar((int)z_mem[0x12], save_file)) == EOF)
     {
-      _handle_save_or_restore_failure(evaluate_result,
+      return _handle_save_or_restore_failure(evaluate_result,
           i18n_libfizmo_ERROR_WRITING_SAVE_FILE, save_file, true);
-      return;
     }
     if ((fsi->writechar((int)z_mem[0x13], save_file)) == EOF)
     {
-      _handle_save_or_restore_failure(evaluate_result,
+      return _handle_save_or_restore_failure(evaluate_result,
           i18n_libfizmo_ERROR_WRITING_SAVE_FILE, save_file, true);
-      return;
     }
     if ((fsi->writechar((int)z_mem[0x14], save_file)) == EOF)
     {
-      _handle_save_or_restore_failure(evaluate_result,
+      return _handle_save_or_restore_failure(evaluate_result,
           i18n_libfizmo_ERROR_WRITING_SAVE_FILE, save_file, true);
-      return;
     }
     if ((fsi->writechar((int)z_mem[0x15], save_file)) == EOF)
     {
-      _handle_save_or_restore_failure(evaluate_result,
+      return _handle_save_or_restore_failure(evaluate_result,
           i18n_libfizmo_ERROR_WRITING_SAVE_FILE, save_file, true);
-      return;
     }
     if ((fsi->writechar((int)z_mem[0x16], save_file)) == EOF)
     {
-      _handle_save_or_restore_failure(evaluate_result,
+      return _handle_save_or_restore_failure(evaluate_result,
           i18n_libfizmo_ERROR_WRITING_SAVE_FILE, save_file, true);
-      return;
     }
     if ((fsi->writechar((int)z_mem[0x17], save_file)) == EOF)
     {
-      _handle_save_or_restore_failure(evaluate_result,
+      return _handle_save_or_restore_failure(evaluate_result,
           i18n_libfizmo_ERROR_WRITING_SAVE_FILE, save_file, true);
-      return;
     }
 
     // Save checksum
     if ((fsi->writechar((int)z_mem[0x1c], save_file)) == EOF)
     {
-      _handle_save_or_restore_failure(evaluate_result,
+      return _handle_save_or_restore_failure(evaluate_result,
           i18n_libfizmo_ERROR_WRITING_SAVE_FILE, save_file, true);
-      return;
     }
     if ((fsi->writechar((int)z_mem[0x1d], save_file)) == EOF)
     {
-      _handle_save_or_restore_failure(evaluate_result,
+      return _handle_save_or_restore_failure(evaluate_result,
           i18n_libfizmo_ERROR_WRITING_SAVE_FILE, save_file, true);
-      return;
     }
 
     // Save initial PC on restore
     if ((fsi->writechar((int)(pc_on_restore >> 16), save_file)) == EOF)
     {
-      _handle_save_or_restore_failure(evaluate_result,
+      return _handle_save_or_restore_failure(evaluate_result,
           i18n_libfizmo_ERROR_WRITING_SAVE_FILE, save_file, true);
-      return;
     }
     if ((fsi->writechar((int)(pc_on_restore >>  8), save_file)) == EOF)
     {
-      _handle_save_or_restore_failure(evaluate_result,
+      return _handle_save_or_restore_failure(evaluate_result,
           i18n_libfizmo_ERROR_WRITING_SAVE_FILE, save_file, true);
-      return;
     }
     if ((fsi->writechar((int)(pc_on_restore      ), save_file)) == EOF)
     {
-      _handle_save_or_restore_failure(evaluate_result,
+      return _handle_save_or_restore_failure(evaluate_result,
           i18n_libfizmo_ERROR_WRITING_SAVE_FILE, save_file, true);
-      return;
     }
 
     if (end_current_chunk(save_file) != 0)
     {
-      _handle_save_or_restore_failure(evaluate_result,
+      return _handle_save_or_restore_failure(evaluate_result,
           i18n_libfizmo_ERROR_WRITING_SAVE_FILE, save_file, true);
-      return;
     }
 
     dynamic_index = z_mem + address;
@@ -654,9 +717,8 @@ void save_game(uint16_t address, uint16_t length, char *filename,
 
       if (start_new_chunk("CMem", save_file) != 0)
       {
-        _handle_save_or_restore_failure(evaluate_result,
+        return _handle_save_or_restore_failure(evaluate_result,
             i18n_libfizmo_ERROR_WRITING_SAVE_FILE, save_file, true);
-        return;
       }
 
       ptr = z_mem + address + length;
@@ -664,11 +726,10 @@ void save_game(uint16_t address, uint16_t length, char *filename,
 
       while (dynamic_index != ptr)
       {
-        if ((data = fsi->getchar(active_z_story->z_story_file)) == EOF)
+        if ((data = fsi->readchar(active_z_story->z_story_file)) == EOF)
         {
-          _handle_save_or_restore_failure(evaluate_result,
+          return _handle_save_or_restore_failure(evaluate_result,
               i18n_libfizmo_ERROR_WRITING_SAVE_FILE, save_file, true);
-          return;
         }
 
         data ^= *dynamic_index;
@@ -689,19 +750,17 @@ void save_game(uint16_t address, uint16_t length, char *filename,
           {
             if (fsi->writechar(0, save_file) == EOF)
             {
-              _handle_save_or_restore_failure(evaluate_result,
+              return _handle_save_or_restore_failure(evaluate_result,
                   i18n_libfizmo_ERROR_WRITING_SAVE_FILE, save_file, true);
-              return;
             }
 
             if (consecutive_zeros > 256)
             {
               if (fsi->writechar(0xff, save_file) == EOF)
               {
-                _handle_save_or_restore_failure(evaluate_result,
+                return _handle_save_or_restore_failure(evaluate_result,
                     i18n_libfizmo_ERROR_WRITING_SAVE_FILE,
                     save_file, true);
-                return;
               }
 
               consecutive_zeros -= 256;
@@ -710,10 +769,9 @@ void save_game(uint16_t address, uint16_t length, char *filename,
             {
               if (fsi->writechar((int)(consecutive_zeros - 1),save_file) == EOF)
               {
-                _handle_save_or_restore_failure(evaluate_result,
+                return _handle_save_or_restore_failure(evaluate_result,
                     i18n_libfizmo_ERROR_WRITING_SAVE_FILE,
                     save_file, true);
-                return;
               }
 
               consecutive_zeros = 0;
@@ -722,10 +780,9 @@ void save_game(uint16_t address, uint16_t length, char *filename,
 
           if (fsi->writechar(data, save_file) == EOF)
           {
-            _handle_save_or_restore_failure(evaluate_result,
+            return _handle_save_or_restore_failure(evaluate_result,
                 i18n_libfizmo_ERROR_WRITING_SAVE_FILE,
                 save_file, true);
-            return;
           }
         }
 
@@ -735,10 +792,9 @@ void save_game(uint16_t address, uint16_t length, char *filename,
 
       if (end_current_chunk(save_file) != 0)
       {
-        _handle_save_or_restore_failure(evaluate_result,
+        return _handle_save_or_restore_failure(evaluate_result,
             i18n_libfizmo_ERROR_WRITING_SAVE_FILE,
             save_file, true);
-        return;
       }
     }
     else
@@ -754,32 +810,29 @@ void save_game(uint16_t address, uint16_t length, char *filename,
       {
         if ((fsi->writechar((int)(*dynamic_index), save_file)) == EOF)
         {
-          _handle_save_or_restore_failure(
+          return _handle_save_or_restore_failure(
               evaluate_result,
               i18n_libfizmo_ERROR_WRITING_SAVE_FILE,
               save_file, true);
-          return;
         }
         dynamic_index++;
       }
 
       if (end_current_chunk(save_file) != 0)
       {
-        _handle_save_or_restore_failure(
+        return _handle_save_or_restore_failure(
             evaluate_result,
             i18n_libfizmo_ERROR_WRITING_SAVE_FILE,
             save_file, true);
-        return;
       }
     }
 
     if (start_new_chunk("Stks", save_file) != 0)
     {
-      _handle_save_or_restore_failure(
+      return _handle_save_or_restore_failure(
           evaluate_result,
           i18n_libfizmo_ERROR_WRITING_SAVE_FILE,
           save_file, true);
-      return;
     }
 
 #ifdef ENABLE_TRACING
@@ -794,27 +847,24 @@ void save_game(uint16_t address, uint16_t length, char *filename,
           number_of_locals_active,
           save_file) != 0)
     {
-      _handle_save_or_restore_failure(evaluate_result,
+      return _handle_save_or_restore_failure(evaluate_result,
           i18n_libfizmo_ERROR_WRITING_SAVE_FILE,
           save_file, true);
-      return;
     }
 
     if (end_current_chunk(save_file) != 0)
     {
-      _handle_save_or_restore_failure(evaluate_result,
+      return _handle_save_or_restore_failure(evaluate_result,
           i18n_libfizmo_ERROR_WRITING_SAVE_FILE,
           save_file, true);
-      return;
     }
 
     if (start_new_chunk("ANNO", save_file) != 0)
     {
-      _handle_save_or_restore_failure(
+      return _handle_save_or_restore_failure(
           evaluate_result,
           i18n_libfizmo_ERROR_WRITING_SAVE_FILE,
           save_file, true);
-      return;
     }
 
     fsi->fileprintf(save_file,
@@ -822,10 +872,9 @@ void save_game(uint16_t address, uint16_t length, char *filename,
 
     if (end_current_chunk(save_file) != 0)
     {
-      _handle_save_or_restore_failure(evaluate_result,
+      return _handle_save_or_restore_failure(evaluate_result,
           i18n_libfizmo_ERROR_WRITING_SAVE_FILE,
           save_file, true);
-      return;
     }
 
 #ifndef DISABLE_OUTPUT_HISTORY
@@ -850,11 +899,10 @@ void save_game(uint16_t address, uint16_t length, char *filename,
 
       if (start_new_chunk("TxHs", save_file) != 0)
       {
-        _handle_save_or_restore_failure(
+        return _handle_save_or_restore_failure(
             evaluate_result,
             i18n_libfizmo_ERROR_WRITING_SAVE_FILE,
             save_file, true);
-        return;
       }
 
       hst_ptr = history->current_paragraph_index;
@@ -865,11 +913,10 @@ void save_game(uint16_t address, uint16_t length, char *filename,
         {
           if (write_four_byte_number(*hst_ptr, save_file) != 0)
           {
-            _handle_save_or_restore_failure(
+            return _handle_save_or_restore_failure(
                 evaluate_result,
                 i18n_libfizmo_ERROR_WRITING_SAVE_FILE,
                 save_file, true);
-            return;
           }
 
           hst_ptr++;
@@ -882,11 +929,10 @@ void save_game(uint16_t address, uint16_t length, char *filename,
       {
         if (write_four_byte_number(*hst_ptr, save_file) != 0)
         {
-          _handle_save_or_restore_failure(
+          return _handle_save_or_restore_failure(
               evaluate_result,
               i18n_libfizmo_ERROR_WRITING_SAVE_FILE,
               save_file, true);
-          return;
         }
 
         hst_ptr++;
@@ -894,25 +940,25 @@ void save_game(uint16_t address, uint16_t length, char *filename,
 
       if (end_current_chunk(save_file) != 0)
       {
-        _handle_save_or_restore_failure(evaluate_result,
+        return _handle_save_or_restore_failure(evaluate_result,
             i18n_libfizmo_ERROR_WRITING_SAVE_FILE,
             save_file, true);
-        return;
       }
     }
 #endif // DISABLE_OUTPUT_HISTORY
 
     if (close_simple_iff_file(save_file) != -0)
     {
-      _handle_save_or_restore_failure(evaluate_result,
+      return _handle_save_or_restore_failure(evaluate_result,
           i18n_libfizmo_ERROR_WRITING_SAVE_FILE,
           save_file, true);
-      return;
     }
   }
 
   if (bool_equal(evaluate_result, true))
     _store_save_or_restore_result(1);
+
+  return 1;
 }
 
 
@@ -986,9 +1032,9 @@ void opcode_save_ext(void)
 }
 
 
-// Returns 0 for failure,  1 for "save succeeded" and 2 for "the game is
-// being restored and is resuming execution again from here, the point where
-// it was saved".
+/* Returns 0 for failure, 2 for successful restore. (These values match
+   the Z-machine @save result codes.)
+   This closes the iff_file. */
 int restore_game_from_stream(uint16_t address, uint16_t length,
     z_file *iff_file, bool evaluate_result)
 {
@@ -1036,17 +1082,17 @@ int restore_game_from_stream(uint16_t address, uint16_t length,
     return _handle_save_or_restore_failure(evaluate_result,
         i18n_libfizmo_ERROR_READING_SAVE_FILE, iff_file, false);
 
-  if (fsi->getchars(release_number, 2, iff_file) != 2)
+  if (fsi->readchars(release_number, 2, iff_file) != 2)
     return _handle_save_or_restore_failure(evaluate_result,
         i18n_libfizmo_COULD_NOT_READ_RELEASE_NUMBER,
         iff_file, false);
 
-  if (fsi->getchars(serial_number, 6, iff_file) != 6)
+  if (fsi->readchars(serial_number, 6, iff_file) != 6)
     return _handle_save_or_restore_failure(evaluate_result,
         i18n_libfizmo_COULD_NOT_READ_SERIAL_NUMBER,
         iff_file, false);
 
-  if (fsi->getchars(checksum, 2, iff_file) != 2)
+  if (fsi->readchars(checksum, 2, iff_file) != 2)
     return _handle_save_or_restore_failure(evaluate_result,
         i18n_libfizmo_COULD_NOT_READ_CHECKSUM, iff_file, false);
 
@@ -1069,7 +1115,7 @@ int restore_game_from_stream(uint16_t address, uint16_t length,
         i18n_libfizmo_RELEASE_NR_SERIAL_NR_OR_CHECKSUM_DOESNT_MATCH,
         iff_file, false);
 
-  if (fsi->getchars(pc_on_restore_data, 3, iff_file) != 3)
+  if (fsi->readchars(pc_on_restore_data, 3, iff_file) != 3)
     return _handle_save_or_restore_failure(evaluate_result,
         i18n_libfizmo_COULD_NOT_READ_RESTORE_PC, iff_file, false);
 
@@ -1136,7 +1182,7 @@ int restore_game_from_stream(uint16_t address, uint16_t length,
     while (bytes_read < chunk_length)
     {
       // Read data from CMem chunk.
-      data = fsi->getchar(iff_file);
+      data = fsi->readchar(iff_file);
       bytes_read++;
 
       if (data == EOF)
@@ -1148,7 +1194,7 @@ int restore_game_from_stream(uint16_t address, uint16_t length,
 
       if (data != 0)
       {
-        data2 = fsi->getchar(active_z_story->z_story_file);
+        data2 = fsi->readchar(active_z_story->z_story_file);
         // Found content difference to original story file.
 
         if (data2 == EOF)
@@ -1171,7 +1217,7 @@ int restore_game_from_stream(uint16_t address, uint16_t length,
       {
         // Found block identical to story file.
 
-        data = fsi->getchar(iff_file);
+        data = fsi->readchar(iff_file);
         bytes_read++;
         if (data == EOF)
         {
@@ -1186,7 +1232,7 @@ int restore_game_from_stream(uint16_t address, uint16_t length,
 
         for (i=0; i<copylength; i++)
         {
-          data2 = fsi->getchar(active_z_story->z_story_file);
+          data2 = fsi->readchar(active_z_story->z_story_file);
           if (data2 == EOF)
           {
             free(restored_story_mem);
@@ -1206,7 +1252,7 @@ int restore_game_from_stream(uint16_t address, uint16_t length,
 
     while (dynamic_index != ptr)
     {
-      data = fsi->getchar(active_z_story->z_story_file);
+      data = fsi->readchar(active_z_story->z_story_file);
       if (data == EOF)
       {
         free(restored_story_mem);
@@ -1245,7 +1291,7 @@ int restore_game_from_stream(uint16_t address, uint16_t length,
     ptr = restored_story_mem + length;
     while (dynamic_index != ptr)
     {
-      data = fsi->getchar(iff_file);
+      data = fsi->readchar(iff_file);
       if (data == EOF)
       {
         free(restored_story_mem);
@@ -1294,7 +1340,7 @@ int restore_game_from_stream(uint16_t address, uint16_t length,
     // Each while iteration processes a single stack frame.
 
     // PC Bits 16-23
-    data = fsi->getchar(iff_file);
+    data = fsi->readchar(iff_file);
     if (data == EOF)
     {
       free(restored_story_mem);
@@ -1305,7 +1351,7 @@ int restore_game_from_stream(uint16_t address, uint16_t length,
     stack_frame_return_pc = (data & 0xff) << 16;
 
     // PC Bits 8-15
-    data = fsi->getchar(iff_file);
+    data = fsi->readchar(iff_file);
     if (data == EOF)
     {
       free(restored_story_mem);
@@ -1316,7 +1362,7 @@ int restore_game_from_stream(uint16_t address, uint16_t length,
     stack_frame_return_pc |= (data & 0xff) << 8;
 
     // PC Bits 0-7
-    data = fsi->getchar(iff_file);
+    data = fsi->readchar(iff_file);
     if (data == EOF)
     {
       free(restored_story_mem);
@@ -1326,7 +1372,7 @@ int restore_game_from_stream(uint16_t address, uint16_t length,
     }
     stack_frame_return_pc |= (data & 0xff);
 
-    data = fsi->getchar(iff_file);
+    data = fsi->readchar(iff_file);
     if (data == EOF)
     {
       free(restored_story_mem);
@@ -1337,7 +1383,7 @@ int restore_game_from_stream(uint16_t address, uint16_t length,
     stack_frame_discard_result = ((data & 0x10) != 0 ? true : false);
     current_stack_frame_nof_locals = (data & 0xf);
 
-    data = fsi->getchar(iff_file);
+    data = fsi->readchar(iff_file);
     if (data == EOF)
     {
       free(restored_story_mem);
@@ -1347,7 +1393,7 @@ int restore_game_from_stream(uint16_t address, uint16_t length,
     }
     stack_frame_result_var = (data & 0xff);
 
-    data = fsi->getchar(iff_file);
+    data = fsi->readchar(iff_file);
     if (data == EOF)
     {
       free(restored_story_mem);
@@ -1357,7 +1403,7 @@ int restore_game_from_stream(uint16_t address, uint16_t length,
     }
     stack_frame_argument_mask = (data & 0xff);
 
-    data = fsi->getchar(iff_file);
+    data = fsi->readchar(iff_file);
     if (data == EOF)
     {
       free(restored_story_mem);
@@ -1367,7 +1413,7 @@ int restore_game_from_stream(uint16_t address, uint16_t length,
     }
     current_stack_frame_nof_functions_stack_words = ((data & 0xff) << 8);
 
-    data = fsi->getchar(iff_file);
+    data = fsi->readchar(iff_file);
     if (data == EOF)
     {
       free(restored_story_mem);
@@ -1402,7 +1448,7 @@ int restore_game_from_stream(uint16_t address, uint16_t length,
     while (i < current_stack_frame_nof_locals
         + current_stack_frame_nof_functions_stack_words)
     {
-      data = fsi->getchar(iff_file);
+      data = fsi->readchar(iff_file);
       if (data == EOF)
       {
         free(restored_story_mem);
@@ -1412,7 +1458,7 @@ int restore_game_from_stream(uint16_t address, uint16_t length,
       }
       stack_word = (data & 0xff) << 8;
 
-      data = fsi->getchar(iff_file);
+      data = fsi->readchar(iff_file);
       if (data == EOF)
       {
         free(restored_story_mem);
@@ -1521,9 +1567,8 @@ int restore_game_from_stream(uint16_t address, uint16_t length,
       "Restored stack: %d locals active, %d, words from active routine.\n",
       number_of_locals_active, stack_words_from_active_routine);
 
-  if (saved_stack->z_stack != NULL)
-    free(saved_stack->z_stack);
-  free(saved_stack);
+  if (saved_stack != NULL)
+    delete_stack_container(saved_stack);
 
   // restored_story_mem has been filled via dynamix_index above so inhibit
   // warning is okay.
@@ -1724,19 +1769,19 @@ bool detect_saved_game(char *file_to_check, char **story_file_to_load)
       return false;
     }
 
-    if (fsi->getchars(release_number_buf, 2, iff_file) != 2)
+    if (fsi->readchars(release_number_buf, 2, iff_file) != 2)
     {
       fsi->closefile(iff_file);
       return false;
     }
 
-    if (fsi->getchars(serial_number, 6, iff_file) != 6)
+    if (fsi->readchars(serial_number, 6, iff_file) != 6)
     {
       fsi->closefile(iff_file);
       return false;
     }
 
-    if (fsi->getchars(checksum_buf, 2, iff_file) != 2)
+    if (fsi->readchars(checksum_buf, 2, iff_file) != 2)
     {
       fsi->closefile(iff_file);
       return false;
