@@ -3,7 +3,7 @@
  *
  * This file is part of fizmo.
  *
- * Copyright (c) 2009-2012 Christoph Ender.
+ * Copyright (c) 2009-2013 Christoph Ender.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -61,7 +61,7 @@
 
 //int stream_active[5] = { 0, 1, 0, 0, 0 };
 bool stream_1_active = true;
-bool stream_2_was_already_active = false;
+bool stream_2_filename_stored = false;
 // stream2 is stored in dynamic memory at z_mem[0x11].
 // stream3 is active when stream_3_current_depth >= -1
 bool stream_4_active = false;
@@ -87,6 +87,11 @@ z_file *input_stream_1 = NULL;
 static bool input_stream_init_underway = false;
 bool input_stream_1_active = false;
 bool input_stream_1_was_already_active = false;
+static z_ucs stream_2_preload_buffer[STREAM_2_PRELOAD_BUFFER_SIZE];
+static z_ucs *stream_2_preload_buffer_index = stream_2_preload_buffer;
+static z_ucs *stream_2_preload_output_start = stream_2_preload_buffer;
+static z_ucs *stream_2_preload_buffer_last_index = (
+    stream_2_preload_buffer + STREAM_2_PRELOAD_BUFFER_SIZE - 1);
 
 // This flag is used for the timed input. Since the verification routine
 // may print some text to the screen, the interpreter must be able to
@@ -94,6 +99,8 @@ bool input_stream_1_was_already_active = false;
 bool stream_output_has_occured = false;
 static bool stream_2_init_underway = false;
 static bool stream_4_was_already_active = false;
+static bool stream_2_wrapping_disabled;
+static int stream2margin;
 
 
 
@@ -105,42 +112,135 @@ void open_streams()
 }
 
 
-/*@-paramuse@*/
-static void stream_2_wrapped_output_destination(z_ucs *z_ucs_output,
+static void stream_2_output_destination(z_ucs *z_ucs_output,
     void *UNUSED(dummy))
 {
-  char buf[128];
-  int len;
-
-  // FIMXE: Re-implement for various output charsets.
-
   if (*z_ucs_output != 0)
   {
-    while (*z_ucs_output != 0)
-    {
-      len = zucs_string_to_utf8_string(buf, &z_ucs_output, 128);
-      fsi->writechars(buf, len-1, stream_2);
-    }
+    fsi->writeucsstring(z_ucs_output, stream_2);
     if (strcmp(get_configuration_value("sync-transcript"), "true") == 0)
       fsi->flushfile(stream_2);
   }
 }
-/*@+paramuse@*/
 
 
-/*
-static void stream_2_do_nothing(void *UNUSED(destination_parameter))
+static void flush_stream_2_buffer_output() {
+  z_ucs *index;
+  z_ucs *newline_index = NULL;
+  z_ucs buf;
+
+  TRACE_LOG("stream_2_preload_flush:\n");
+
+  index
+    = stream_2_preload_buffer_index == stream_2_preload_buffer
+    ? stream_2_preload_buffer_last_index - 1
+    : stream_2_preload_buffer_index -1;
+
+  while (index != stream_2_preload_output_start) {
+    if (*index == '\n') {
+      newline_index = index;
+      break;
+    }
+    if (index == stream_2_preload_buffer) {
+      index = stream_2_preload_buffer_last_index;
+    }
+    index--;
+  }
+
+  if (newline_index == NULL) {
+    index = stream_2_preload_output_start + (STREAM_2_PRELOAD_BUFFER_SIZE / 2);
+    if (index >= stream_2_preload_buffer_last_index) {
+      index = stream_2_preload_buffer
+        + (index - stream_2_preload_buffer_last_index);
+    }
+  }
+  else {
+    if ((index = newline_index + 1) == stream_2_preload_buffer_last_index) {
+      index = stream_2_preload_buffer;
+    }
+  }
+
+  if (stream_2_preload_output_start > index) {
+    TRACE_LOG("stream_2_preload in wrap-around.\n");
+    TRACE_LOG("flushs2: \"");
+    TRACE_LOG_Z_UCS(stream_2_preload_output_start);
+    TRACE_LOG("\".\n");
+    // In wrap-around
+    stream_2_output_destination(stream_2_preload_output_start, NULL);
+    stream_2_preload_output_start = stream_2_preload_buffer;
+  }
+
+  buf = *index;
+  *index = 0;
+  TRACE_LOG("flushs2:\"");
+  TRACE_LOG_Z_UCS(stream_2_preload_output_start);
+  TRACE_LOG("\".\n");
+  stream_2_output_destination(stream_2_preload_output_start, NULL);
+  *index = buf;
+  stream_2_preload_output_start = index;
+}
+
+
+static void stream_2_buffer_output(z_ucs *z_ucs_output)
 {
-};
+  //z_ucs buf;
+
+  while (*z_ucs_output != 0) {
+    TRACE_LOG("stream_2_preload_buffer_index: %x\n",
+        stream_2_preload_buffer_index);
+    TRACE_LOG("stream_2_preload_buffer_last_index: %x\n",
+        stream_2_preload_buffer_last_index);
+    TRACE_LOG("stream_2_preload_output_start: %x\n",
+        stream_2_preload_output_start);
+
+    if (stream_2_preload_buffer_index == stream_2_preload_buffer_last_index) {
+      // output-start following buffer_index in wrap-around:
+      if (stream_2_preload_output_start == stream_2_preload_buffer) {
+        flush_stream_2_buffer_output();
+      }
+      stream_2_preload_buffer_index = stream_2_preload_buffer;
+    }
+
+    // output-start following buffer_index as non-wrap-around:
+    if (stream_2_preload_buffer_index + 1 == stream_2_preload_output_start) {
+      flush_stream_2_buffer_output();
+    }
+
+    TRACE_LOG("Writing %c to %x.\n",
+        *z_ucs_output, stream_2_preload_buffer_index);
+    *stream_2_preload_buffer_index = *z_ucs_output;
+    stream_2_preload_buffer_index++;
+    z_ucs_output++;
+  }
+}
 
 
-static struct wordwrap_target streams_wordwrap_target =
-{
-  &stream_2_wrapped_output_destination,
-  &stream_2_do_nothing,
-  &stream_2_do_nothing
-};
-*/
+int stream_2_buffer_remove_chars(size_t nof_chars_to_remove) {
+  while (nof_chars_to_remove > 0) {
+
+    if (stream_2_preload_buffer_index == stream_2_preload_output_start) {
+      TRACE_LOG("remove-s2 hit output start.\n");
+      return -1;
+    }
+
+    if (stream_2_preload_buffer_index == stream_2_preload_buffer) {
+      stream_2_preload_buffer_index = stream_2_preload_buffer_last_index - 1;
+    }
+    else {
+      stream_2_preload_buffer_index--;
+    }
+    TRACE_LOG("Removing %c.\n", *stream_2_preload_buffer_index);
+    TRACE_LOG("stream_2_preload_buffer_index: %x\n",
+        stream_2_preload_buffer_index);
+    TRACE_LOG("stream_2_preload_buffer_last_index: %x\n",
+        stream_2_preload_buffer_last_index);
+    TRACE_LOG("stream_2_preload_output_start: %x\n",
+        stream_2_preload_output_start);
+
+    nof_chars_to_remove--;
+  }
+  return 0;
+}
 
 
 void init_streams()
@@ -149,7 +249,9 @@ void init_streams()
   size_t bytes_required;
   char *stream_2_line_width = get_configuration_value("stream-2-line-width");
   char *stream_2_left_margin = get_configuration_value("stream-2-left-margin");
-  int stream2width, stream2margin;
+  stream_2_wrapping_disabled
+    = (strcmp(get_configuration_value("disable-stream-2-wrap"), "true") == 0);
+  int stream2width;
 
   stream2width
     = stream_2_line_width != NULL       
@@ -161,18 +263,22 @@ void init_streams()
     ? atoi(stream_2_left_margin)
     : DEFAULT_STREAM_2_LEFT_PADDING;
 
-  stream_2_wrapper = wordwrap_new_wrapper(
-      stream2width,
-      &stream_2_wrapped_output_destination,
-      NULL,
-      true,
-      stream2margin,
-      (strcmp(get_configuration_value("sync-transcript"), "true") == 0
-       ? true
-       : false),
-      (strcmp(get_configuration_value("disable-stream-2-hyphenation"),"true")==0
-       ? false
-       : true));
+  if (stream_2_wrapping_disabled == false)
+    stream_2_wrapper = wordwrap_new_wrapper(
+        stream2width,
+        &stream_2_output_destination,
+        NULL,
+        true,
+        stream2margin,
+        (strcmp(get_configuration_value("sync-transcript"), "true") == 0
+         ? true
+         : false),
+        (strcmp(get_configuration_value("disable-stream-2-hyphenation")
+                ,"true") == 0
+         ? false
+         : true));
+
+  *stream_2_preload_buffer_last_index = 0;
 
   if (strcmp(get_configuration_value("start-script-when-story-starts"),
         "true") == 0)
@@ -193,10 +299,10 @@ void init_streams()
     }
 
     strcpy(stream_2_filename, src);
-    stream_2_was_already_active = true;
+    stream_2_filename_stored = true;
   }
   else
-    src = fizmo_strdup(DEFAULT_TRANSCRIPT_FILE_NAME);
+    src = DEFAULT_TRANSCRIPT_FILE_NAME;
 
   (void)latin1_string_to_zucs_string(
       last_script_filename,
@@ -222,7 +328,7 @@ void init_streams()
     input_stream_1_was_already_active = true;
   }
   else
-    src = fizmo_strdup(DEFAULT_INPUT_COMMAND_FILE_NAME);
+    src = DEFAULT_INPUT_COMMAND_FILE_NAME;
 
   (void)latin1_string_to_zucs_string(
       last_input_stream_filename,
@@ -243,7 +349,7 @@ void init_streams()
     stream_4_was_already_active = true;
   }
   else
-    src = fizmo_strdup(DEFAULT_RECORD_COMMAND_FILE_NAME);
+    src = DEFAULT_RECORD_COMMAND_FILE_NAME;
 
   (void)latin1_string_to_zucs_string(
       last_stream_4_filename,
@@ -255,16 +361,40 @@ void init_streams()
     input_stream_1_active = true;
 }
 
+z_file *get_stream_2(void)
+{
+  return stream_2;
+}
+
+
+/* Accept a new open stream as the current stream_2. The previous stream_2
+   is closed. Pass in NULL to just close the previous stream_2.
+   The interpreter's transcript bit is set appropriately.
+ */
+void restore_stream_2(z_file *str)
+{
+  if (stream_2) {
+    (void)fsi->closefile(stream_2);
+    stream_2 = NULL;
+    z_mem[0x11] &= 0xfe;
+  }
+  if (str) {
+    stream_2 = str;
+    z_mem[0x11] |= 1;
+  }
+}
 
 void ask_for_input_stream_filename(void)
 {
-  z_ucs *current_line;
   bool stream_1_active_buf;
   int16_t input_length;
   int16_t i;
   size_t bytes_required;
   z_ucs *ptr;
   int len;
+#ifndef DISABLE_OUTPUT_HISTORY
+  z_ucs *current_line;
+#endif // DISABLE_OUTPUT_HISTORY
 
   input_stream_init_underway = true;
 
@@ -284,6 +414,7 @@ void ask_for_input_stream_filename(void)
 
   (void)streams_z_ucs_output(last_input_stream_filename);
   len = z_ucs_len(last_input_stream_filename);
+  stream_2_remove_chars(len);
 #ifndef DISABLE_OUTPUT_HISTORY
   remove_chars_from_history(outputhistory[active_window_number], len);
 #endif /* DISABLE_OUTPUT_HISTORY */
@@ -381,11 +512,13 @@ void ask_for_input_stream_filename(void)
 
   TRACE_LOG("Converted filename: '%s'.\n", stream_2_filename);
 
+#ifndef DISABLE_OUTPUT_HISTORY
   if (current_line != NULL)
   {
     (void)streams_z_ucs_output(current_line);
     free(current_line);
   }
+#endif // DISABLE_OUTPUT_HISTORY
 
   input_stream_init_underway = false;
 }
@@ -393,20 +526,20 @@ void ask_for_input_stream_filename(void)
 
 void ask_for_stream2_filename()
 {
-  z_ucs *current_line;
   bool stream_1_active_buf;
   int16_t input_length;
   int16_t i;
   size_t bytes_required;
   z_ucs *ptr;
   int len;
+#ifndef DISABLE_OUTPUT_HISTORY
+  z_ucs *current_line;
+#endif // DISABLE_OUTPUT_HISTORY
 
   stream_2_init_underway = true;
 
 #ifndef DISABLE_OUTPUT_HISTORY
   current_line = get_current_line(outputhistory[active_window_number]);
-#else
-  current_line = NULL;
 #endif /* DISABLE_OUTPUT_HISTORY */
 
   stream_1_active_buf = stream_1_active;
@@ -425,6 +558,7 @@ void ask_for_stream2_filename()
         "streams_z_ucs_output");
 
   len = z_ucs_len(last_script_filename);
+  stream_2_remove_chars(len);
 #ifndef DISABLE_OUTPUT_HISTORY
   remove_chars_from_history(outputhistory[active_window_number], len);
 #endif /* DISABLE_OUTPUT_HISTORY */
@@ -522,66 +656,125 @@ void ask_for_stream2_filename()
 
   TRACE_LOG("Converted filename: '%s'.\n", stream_2_filename);
 
+#ifndef DISABLE_OUTPUT_HISTORY
   if (current_line != NULL)
   {
     (void)streams_z_ucs_output(current_line);
     free(current_line);
   }
-  stream_2_was_already_active = true;
+#endif // DISABLE_OUTPUT_HISTORY
+
+  stream_2_filename_stored = true;
   stream_2_init_underway = false;
 }
 
 
+static void stream_2_output_write(z_ucs *z_ucs_output)
+{
+  if (stream_2_wrapping_disabled == true)
+    stream_2_buffer_output(z_ucs_output);
+  else
+    wordwrap_wrap_z_ucs(stream_2_wrapper, z_ucs_output);
+}
+
+
+static void stream_2_print_header()
+{
+  z_ucs dashes[] = { '-', '-', '-', '\n', '\n', 0 };
+
+  stream_2_output_write(z_ucs_newline_string);
+  stream_2_output_write(dashes);
+}
+
+
+void stream_2_remove_chars(size_t nof_chars_to_remove) {
+  if (stream_2 != NULL) {
+    if (stream_2_wrapping_disabled == true) {
+      stream_2_buffer_remove_chars(nof_chars_to_remove);
+    }
+    else {
+      wordwrap_remove_chars(stream_2_wrapper, nof_chars_to_remove);
+    }
+  }
+}
+
+
 static void stream_2_output(z_ucs *z_ucs_output)
 {
+  int return_code;
+  z_file *transcript_stream = NULL;
+
   if (
       (active_interface == NULL)
       ||
       (active_window_number != 0)
+      ||
+      (stream_2_init_underway == true)
      )
     return ;
 
-  if (
-      (stream_2_was_already_active == false)
-      &&
-      (stream_2_init_underway == false)
-     )
+  if (stream_2 == NULL)
   {
-    ask_for_stream2_filename();
-  }
-
-  if (stream_2_init_underway == false)
-  {
-    if (stream_2 == NULL)
+    if (stream_2_filename_stored == false)
     {
-      TRACE_LOG("Opening script-file '%s' for writing.\n", stream_2_filename);
-      stream_2 = fsi->openfile(
-          stream_2_filename, FILETYPE_TRANSCRIPT, FILEACCESS_APPEND);
-      fsi->writechar('\n', stream_2);
-      wordwrap_output_left_padding(stream_2_wrapper);
-      fsi->writestring("---\n\n", stream_2);
-      wordwrap_output_left_padding(stream_2_wrapper);
-      if (strcmp(get_configuration_value("sync-transcript"), "true") == 0)
-        fsi->flushfile(stream_2);
-    }
+      return_code = active_interface->prompt_for_filename(
+          "transcript",
+          &transcript_stream,
+          NULL,
+          FILETYPE_TRANSCRIPT,
+          FILEACCESS_APPEND);
 
-    if (bool_equal(lower_window_buffering_active, true))
-    {
-      if (script_wrapper_active == false)
+      if (return_code == -3)
       {
-        script_wrapper_active = true;
+        // No support for "prompt_for_filename" in screen_interface, default
+        // to built-in method.
+        ask_for_stream2_filename();
+        // This method will not open a file, but instead store the filename in
+        //"stream_2_filename".
+        TRACE_LOG("Opening script-file '%s' for writing.\n", stream_2_filename);
+        stream_2 = fsi->openfile(
+            stream_2_filename, FILETYPE_TRANSCRIPT, FILEACCESS_APPEND);
+      }
+      else if (return_code < 0)
+      {
+        /* The user cancelled out. We'll have to silently turn off stream 2.
+           Not the best option, but the best option I can see how to
+           do. */
+        z_mem[0x11] &= 0xfe;
+        return;
+      }
+      else
+      {
+        stream_2 = transcript_stream;
       }
-      wordwrap_wrap_z_ucs(stream_2_wrapper, z_ucs_output);
     }
     else
     {
-      if (bool_equal(script_wrapper_active, true))
-      {
-        wordwrap_flush_output(stream_2_wrapper);
-        script_wrapper_active = false;
-      }
-      stream_2_wrapped_output_destination(z_ucs_output, NULL);
+      stream_2 = fsi->openfile(
+          stream_2_filename, FILETYPE_TRANSCRIPT, FILEACCESS_APPEND);
     }
+
+    stream_2_print_header();
+  }
+
+  if (bool_equal(lower_window_buffering_active, true))
+  {
+    if (script_wrapper_active == false)
+    {
+      flush_stream_2_buffer_output();
+      script_wrapper_active = true;
+    }
+    stream_2_output_write(z_ucs_output);
+  }
+  else
+  {
+    if (bool_equal(script_wrapper_active, true))
+    {
+      if (stream_2_wrapping_disabled == false)
+        wordwrap_flush_output(stream_2_wrapper);
+      script_wrapper_active = false;
+    }
+    stream_2_buffer_output(z_ucs_output);
   }
 }
 
@@ -592,13 +785,16 @@ void ask_for_stream4_filename_if_required(bool dont_output_current_line)
 void ask_for_stream4_filename_if_required(bool UNUSED(dont_output_current_line))
 #endif // DISABLE_OUTPUT_HISTORY
 {
-  z_ucs *current_line = NULL;
   bool stream_1_active_buf;
   int16_t input_length;
   int16_t i;
   size_t bytes_required;
   z_ucs *ptr;
   int len;
+  int return_code;
+#ifndef DISABLE_OUTPUT_HISTORY
+  z_ucs *current_line = NULL;
+#endif // DISABLE_OUTPUT_HISTORY
 
   if (
       (bool_equal(stream_4_active, true))
@@ -618,116 +814,141 @@ void ask_for_stream4_filename_if_required(bool UNUSED(dont_output_current_line))
     }
 #endif /* DISABLE_OUTPUT_HISTORY */
 
-    stream_1_active_buf = stream_1_active;
-    stream_1_active = true;
+    return_code = active_interface->prompt_for_filename(
+        "transcript",
+        &stream_4,
+        NULL,
+        FILETYPE_INPUTRECORD,
+        FILEACCESS_APPEND);
 
-    do
+    if (return_code >= 0)
     {
-      (void)streams_latin1_output("\n");
+      // Success.
+      stream_4_was_already_active = true;
+      stream_4_init_underway = false;
+    }
+    else if ( (return_code == -1) || (return_code == -2) )
+    {
+      stream_4_active = false;
+    }
+    else if (return_code == -3)
+    {
+      // No support for UI-specific filename prompt.
 
-      (void)i18n_translate(
-          libfizmo_module_name,
-          i18n_libfizmo_PLEASE_ENTER_NAME_FOR_COMMANDFILE);
+      stream_1_active_buf = stream_1_active;
+      stream_1_active = true;
 
-      (void)streams_latin1_output("\n>");
+      do
+      {
+        (void)streams_latin1_output("\n");
 
-      (void)streams_z_ucs_output(last_stream_4_filename);
+        (void)i18n_translate(
+            libfizmo_module_name,
+            i18n_libfizmo_PLEASE_ENTER_NAME_FOR_COMMANDFILE);
 
-      len = z_ucs_len(last_stream_4_filename);
+        (void)streams_latin1_output("\n>");
+
+        (void)streams_z_ucs_output(last_stream_4_filename);
+
+        len = z_ucs_len(last_stream_4_filename);
+        stream_2_remove_chars(len);
 #ifndef DISABLE_OUTPUT_HISTORY
-      remove_chars_from_history(outputhistory[active_window_number], len);
+        remove_chars_from_history(outputhistory[active_window_number], len);
 #endif /* DISABLE_OUTPUT_HISTORY */
 
-      for (i=0; i<len; i++)
+        for (i=0; i<len; i++)
+        {
+          current_filename_buffer[i]
+            = unicode_char_to_zscii_input_char(last_stream_4_filename[i]);
+        }
+
+        input_length
+          = active_interface->read_line(
+              (zscii*)current_filename_buffer,
+              MAXIMUM_SCRIPT_FILE_NAME_LENGTH,
+              0,
+              0,
+              len,
+              NULL,
+              true,
+              false);
+
+        if (input_length == 0)
+        {
+          if (streams_latin1_output("\n") != 0)
+            i18n_translate_and_exit(
+                libfizmo_module_name,
+                i18n_libfizmo_FUNCTION_CALL_P0S_ABORTED_DUE_TO_ERROR,
+                -0x0100,
+                "streams_latin1_output");
+
+          if (i18n_translate(
+                libfizmo_module_name,
+                i18n_libfizmo_FILENAME_MUST_NOT_BE_EMPTY) == (size_t)-1)
+            i18n_translate_and_exit(
+                libfizmo_module_name,
+                i18n_libfizmo_FUNCTION_CALL_P0S_ABORTED_DUE_TO_ERROR,
+                -0x0100,
+                "i18n_translate");
+
+          if (streams_latin1_output("\n") != 0)
+            i18n_translate_and_exit(
+                libfizmo_module_name,
+                i18n_libfizmo_FUNCTION_CALL_P0S_ABORTED_DUE_TO_ERROR,
+                -0x0100,
+                "streams_latin1_output");
+        }
+      }
+      while (input_length == 0);
+
+      stream_1_active = stream_1_active_buf;
+
+      for (i=0; i<input_length; i++)
+        last_stream_4_filename[i]
+          = zscii_input_char_to_z_ucs(current_filename_buffer[i]);
+      last_stream_4_filename[i] = 0;
+
+      (void)streams_z_ucs_output(last_stream_4_filename);
+      (void)streams_latin1_output("\n\n");
+
+      TRACE_LOG("From ZSCII translated filename: \"");
+      TRACE_LOG_Z_UCS(last_stream_4_filename);
+      TRACE_LOG("\".\n");
+
+      ptr = last_stream_4_filename;
+      bytes_required = (size_t)zucs_string_to_utf8_string(NULL, &ptr, 0);
+
+      if (
+          (stream_4_filename == NULL)
+          ||
+          (bytes_required > stream_4_filename_size)
+         )
       {
-        current_filename_buffer[i]
-          = unicode_char_to_zscii_input_char(last_stream_4_filename[i]);
+        TRACE_LOG("(Re-)allocating %zd bytes.\n", bytes_required);
+
+        stream_4_filename
+          = (char*)fizmo_realloc(stream_4_filename, bytes_required);
+
+        stream_4_filename_size = bytes_required;
       }
 
-      input_length
-        = active_interface->read_line(
-            (zscii*)current_filename_buffer,
-            MAXIMUM_SCRIPT_FILE_NAME_LENGTH,
-            0,
-            0,
-            len,
-            NULL,
-            true,
-            false);
+      // FIXME: Charsets may differ on operating systems.
+      ptr = last_stream_4_filename;
+      (void)zucs_string_to_utf8_string(
+          stream_4_filename,
+          &ptr,
+          bytes_required);
 
-      if (input_length == 0)
+#ifndef DISABLE_OUTPUT_HISTORY
+      if (current_line != NULL)
       {
-        if (streams_latin1_output("\n") != 0)
-          i18n_translate_and_exit(
-              libfizmo_module_name,
-              i18n_libfizmo_FUNCTION_CALL_P0S_ABORTED_DUE_TO_ERROR,
-              -0x0100,
-              "streams_latin1_output");
-
-        if (i18n_translate(
-              libfizmo_module_name,
-              i18n_libfizmo_FILENAME_MUST_NOT_BE_EMPTY) == (size_t)-1)
-          i18n_translate_and_exit(
-              libfizmo_module_name,
-              i18n_libfizmo_FUNCTION_CALL_P0S_ABORTED_DUE_TO_ERROR,
-              -0x0100,
-              "i18n_translate");
-
-        if (streams_latin1_output("\n") != 0)
-          i18n_translate_and_exit(
-              libfizmo_module_name,
-              i18n_libfizmo_FUNCTION_CALL_P0S_ABORTED_DUE_TO_ERROR,
-              -0x0100,
-              "streams_latin1_output");
+        (void)streams_z_ucs_output(current_line);
+        free(current_line);
       }
+#endif // DISABLE_OUTPUT_HISTORY
+      stream_4_was_already_active = true;
+      stream_4_init_underway = false;
     }
-    while (input_length == 0);
-
-    stream_1_active = stream_1_active_buf;
-
-    for (i=0; i<input_length; i++)
-      last_stream_4_filename[i]
-        = zscii_input_char_to_z_ucs(current_filename_buffer[i]);
-    last_stream_4_filename[i] = 0;
-
-    (void)streams_z_ucs_output(last_stream_4_filename);
-    (void)streams_latin1_output("\n\n");
-
-    TRACE_LOG("From ZSCII translated filename: \"");
-    TRACE_LOG_Z_UCS(last_stream_4_filename);
-    TRACE_LOG("\".\n");
-
-    ptr = last_stream_4_filename;
-    bytes_required = (size_t)zucs_string_to_utf8_string(NULL, &ptr, 0);
-
-    if (
-        (stream_4_filename == NULL)
-        ||
-        (bytes_required > stream_4_filename_size)
-       )
-    {
-      TRACE_LOG("(Re-)allocating %zd bytes.\n", bytes_required);
-
-      stream_4_filename
-        = (char*)fizmo_realloc(stream_4_filename, bytes_required);
-
-      stream_4_filename_size = bytes_required;
-    }
-
-    // FIXME: Charsets may differ on operating systems.
-    ptr = last_stream_4_filename;
-    (void)zucs_string_to_utf8_string(
-        stream_4_filename,
-        &ptr,
-        bytes_required);
-
-    if (current_line != NULL)
-    {
-      (void)streams_z_ucs_output(current_line);
-      free(current_line);
-    }
-    stream_4_was_already_active = true;
-    stream_4_init_underway = false;
   }
 }
 
@@ -743,7 +964,7 @@ void stream_4_latin1_output(char *latin1_output)
 
   // We'll ask at this point for the filename, since asking directly when
   // the OUTPUT_STREAM opcode is processed might garble then screen output.
-  // So we'll just wair until the user has finished with the input and ask
+  // So we'll just wait until the user has finished with the input and ask
   // for the filename to save to once he's finished.
   ask_for_stream4_filename_if_required(false);
 
@@ -779,7 +1000,10 @@ static void close_script_file()
   {
     TRACE_LOG("Closing script-file.\n");
 
-    wordwrap_flush_output(stream_2_wrapper);
+    if (stream_2_wrapping_disabled == false)
+      wordwrap_flush_output(stream_2_wrapper);
+    else
+      flush_stream_2_buffer_output();
     fsi->writechar('\n', stream_2);
     (void)fsi->closefile(stream_2);
     stream_2 = NULL;
@@ -1088,26 +1312,12 @@ static int _streams_z_ucs_output(z_ucs *z_ucs_output, bool is_user_input)
 
       if (bool_equal(stream_1_active, true))
       {
-        if (active_interface != NULL)
-        {
-          active_interface->z_ucs_output(converted_output);
-        }
-
 #ifndef DISABLE_OUTPUT_HISTORY
         if (active_window_number == 0)
         {
           // outputhistory[0] is always defined if DISABLE_OUTPUT_HISTORY is not
           // defined (see fizmo.c).
 
-          if (lower_window_style != current_style)
-          {
-            store_metadata_in_history(
-                outputhistory[0],
-                HISTORY_METADATA_TYPE_STYLE,
-                current_style);
-            lower_window_style = current_style;
-          }
-
           if (
               (lower_window_foreground_colour != current_foreground_colour)
               ||
@@ -1125,6 +1335,24 @@ static int _streams_z_ucs_output(z_ucs *z_ucs_output, bool is_user_input)
           }
         }
 
+        if ( (active_z_story != NULL) && (z_mem[0x11] & 2) ) {
+          current_resulting_font = Z_FONT_COURIER_FIXED_PITCH;
+          //printf("force fixed on: %d\n", current_resulting_font);
+        }
+        else {
+          current_resulting_font = current_font;
+          //printf("force fixed off: %d\n", current_resulting_font);
+        }
+
+        if (lower_window_font != current_resulting_font) {
+          store_metadata_in_history(
+              outputhistory[0],
+              HISTORY_METADATA_TYPE_FONT,
+              current_resulting_font);
+          lower_window_font = current_resulting_font;
+          active_interface->set_font(current_resulting_font);
+        }
+
         if (
             (
              (ver == 6)
@@ -1168,6 +1396,10 @@ static int _streams_z_ucs_output(z_ucs *z_ucs_output, bool is_user_input)
               upper_window_buffer, converted_output);
         }
 #endif /* DISABLE_BLOCKBUFFER */
+
+        if (active_interface != NULL) {
+          active_interface->z_ucs_output(converted_output);
+        }
       }
 
       if (font_conversion_active == false)
@@ -1324,15 +1556,17 @@ void close_streams(z_ucs *error_message)
   if (stream_2 != NULL)
     close_script_file();
 
+  if (stream_2_wrapper != NULL)
+  {
+    wordwrap_destroy_wrapper(stream_2_wrapper);
+    stream_2_wrapper = NULL;
+  }
+
   if (stream_4 != NULL)
   {
     (void)fsi->closefile(stream_4);
     stream_4 = NULL;
   }
-
-#ifdef ENABLE_TRACING
-  turn_off_trace();
-#endif /* ENABLE_TRACING */
 }
 
 
