@@ -3,7 +3,7 @@
  *
  * This file is part of fizmo.
  *
- * Copyright (c) 2009-2012 Christoph Ender.
+ * Copyright (c) 2009-2013 Christoph Ender.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -51,7 +51,7 @@
 #include "streams.h"
 #include "history.h"
 #include "fizmo.h"
-#include "math.h"
+#include "mathemat.h"
 #include "stack.h"
 #include "config.h"
 #include "output.h"
@@ -68,15 +68,14 @@
 #include "cmd_hst.h"
 #endif /* DISABLE_COMMAND_HISTORY */
 
-
-static uint16_t zchar_storage_word;
-static uint16_t zchar_storage_symbols_stored;
-static uint8_t *zchar_storage_output;
-static uint8_t *zchar_storage_output_index;
-static uint8_t *zchar_storage_index_behind;
-static uint8_t zchar_storage_word_index;
+static uint16_t zchar_storage_word = 0;
+static uint16_t zchar_storage_symbols_stored = 0;
+static uint8_t *zchar_storage_output = NULL;
+static uint8_t *zchar_storage_output_index = NULL;
+static uint8_t *zchar_storage_index_behind = NULL;
+static uint8_t zchar_storage_word_index = 0;
 static size_t interpreter_command_buffer_size = 0;
-/*@only@*/ /*@null@*/ z_ucs *interpreter_command_buffer = NULL;
+z_ucs *interpreter_command_buffer = NULL;
 
 // The following two variables will keep track of the src pointer in
 // the various abbreviation depths. In case we return from processing
@@ -94,11 +93,6 @@ static uint32_t number_of_commands = 0;
 static uint8_t first_word_found;
 static z_ucs z_ucs_output_buffer[Z_UCS_OUTPUT_BUFFER_SIZE];
 
-static uint8_t dictionary_entry_length;
-static int16_t number_of_dictionary_entries;
-static bool dictionary_is_unsorted;
-static uint8_t *dictionary_start;
-
 z_ucs z_ucs_newline_string[] = { Z_UCS_NEWLINE, 0 };
 
 #ifndef DISABLE_PREFIX_COMMANDS
@@ -179,8 +173,16 @@ static void zchar_storage_start(uint8_t *output, uint16_t output_buffer_size)
 //   * -1 in case the given five_bits could not be stored. These will be lost.
 static int zchar_storage_write(unsigned int five_bits)
 {
+  if (zchar_storage_output_index >= zchar_storage_index_behind)
+  {
+    TRACE_LOG("zchar_storage_write ran out of space.\n");
+    return -1;
+  }
+
   five_bits &= 0x1f;
 
+  TRACE_LOG("zchar: Storing five bits %x.\n", five_bits);
+
   if (zchar_storage_word_index == 0)
     zchar_storage_word |= (five_bits << 10);
   else if (zchar_storage_word_index == 1)
@@ -192,8 +194,10 @@ static int zchar_storage_write(unsigned int five_bits)
 
   if (++zchar_storage_word_index == 3)
   {
-    if (zchar_storage_output_index == zchar_storage_index_behind)
-      return -1;
+    TRACE_LOG("zchar: storing word %x at %x, end at %x.\n",
+        zchar_storage_word,
+        zchar_storage_output_index,
+        zchar_storage_index_behind);
 
     store_word(zchar_storage_output_index, zchar_storage_word);
     zchar_storage_output_index += 2;
@@ -201,10 +205,16 @@ static int zchar_storage_write(unsigned int five_bits)
     zchar_storage_word_index = 0;
   }
 
-  if (zchar_storage_output_index == zchar_storage_index_behind)
+  if (zchar_storage_output_index >= zchar_storage_index_behind)
+  {
+    TRACE_LOG("No more space remains for zchar-storage.\n");
     return 1;
+  }
   else
+  {
+    TRACE_LOG("Space remains for zchar-storage.\n");
     return 0;
+  }
 }
 
 
@@ -230,34 +240,73 @@ void zchar_storage_reset()
 }
 
 
+static void zchar_storage_clear()
+{
+  zchar_storage_word = 0;
+  zchar_storage_output = NULL;
+  zchar_storage_output_index = NULL;
+  zchar_storage_index_behind = NULL;
+  zchar_storage_word_index = 0;
+  zchar_storage_symbols_stored = 0;
+}
+
+
 // This function will finish the current Z-Char-String in production by
 // setting bit 15 on the last word, padding it with 5s if required and
 // storing it in the output buffer.
-static void zchar_storage_finish()
+// Return value:
+// -0 In case there was still enough space and string finalization succeeded.
+// -1 in case there was not enough buffer storage to write the string.
+static int zchar_storage_finish()
 {
+  TRACE_LOG("zchar_storage_word_index: %d.\n", zchar_storage_word_index);
   if (zchar_storage_word_index == 0)
+  {
     // In case we arrive here, there's no output in the current word. We'll
     // check if there's any output already stored in the output buffer.
     if (zchar_storage_output_index != zchar_storage_output)
+    {
       // In case there's some output, write the string termination to the
       // last word of the output.
+      TRACE_LOG("Successfully finalized full word.");
       *(zchar_storage_output_index-2) |= 0x80;
+    }
     else
+    {
       // In case the buffer is completely empty, write an empty word
       // padded with 5s and terminated with bit 15 set.
-      store_word(zchar_storage_output, 0x94a5);
+      if (zchar_storage_output_index >= zchar_storage_index_behind)
+      {
+        TRACE_LOG("zchar_storage_finish ran out of space.\n");
+        return -1;
+      }
 
+      TRACE_LOG("zchar: storing word %x at %x.\n",
+          zchar_storage_word, 0x94a5);
+      store_word(zchar_storage_output, 0x94a5);
+    }
+  }
   else
   {
     // In case the current word has some content, we'll just have pad it
     // up with 5s and store it.
+    if (zchar_storage_output_index >= zchar_storage_index_behind)
+    {
+      TRACE_LOG("zchar_storage_finish ran out of space.\n");
+      return -1;
+    }
+
     if (zchar_storage_word_index == 1)
       zchar_storage_word |= (5 << 5);
 
     zchar_storage_word |= 0x8005;
 
+    TRACE_LOG("zchar: storing word %x at %x.\n",
+        zchar_storage_word, zchar_storage_output_index);
     store_word(zchar_storage_output_index, zchar_storage_word);
   }
+
+  return 0;
 }
 
 
@@ -444,11 +493,17 @@ static uint8_t locate_dictionary_entry(
     uint8_t **parse_position,
     uint8_t word_position,
     uint8_t token_length,
+    uint8_t *dictionary_start,
+    int16_t number_of_dictionary_entries,
+    uint8_t dictionary_entry_length,
+    bool dictionary_is_unsorted,
     bool dont_write_unrecognized_words_to_parse_buffer)
 {
   uint8_t *dictionary_index = dictionary_start;
   uint16_t i,j;
   uint16_t word_length;
+  int16_t word_found_at_index = -1;
+  int search_mid_index, search_end_index;
 
   // "Dictionarize" entry.
   if (ver >= 4)
@@ -486,40 +541,76 @@ static uint8_t locate_dictionary_entry(
     }
   }
 
-  // FIXME: Improve searching (use "dictionary_is_unsorted").
-  for (i=0; i<number_of_dictionary_entries; i++)
+  TRACE_LOG("dict-start: %x\n", dictionary_start - z_mem);
+
+  if (dictionary_is_unsorted == true)
   {
-    /*
-    TRACE_LOG("%d (at %x): %x%x%x%x / %x%x%x%x\n",
-        i,
-        dictionary_index - z_mem,
-        text_buffer[0],
-        text_buffer[1],
-        text_buffer[2],
-        text_buffer[3],
-        dictionary_index[0],
-        dictionary_index[1],
-        dictionary_index[2],
-        dictionary_index[3]);
-    */
+    for (i=0; i<number_of_dictionary_entries; i++)
+    {
+      for (j=0; j<word_length; j++) {
+        TRACE_LOG("%d/%x: %x %x\n",
+            j, dictionary_index-z_mem, text_buffer[j], dictionary_index[j]);
+        if (text_buffer[j] != dictionary_index[j])
+          break;
+      }
 
-    for (j=0; j<word_length; j++)
-      if (text_buffer[j] != dictionary_index[j])
+      if (j == word_length)
+        word_found_at_index = i;
+
+      dictionary_index += dictionary_entry_length;
+    }
+  }
+  else
+  {
+    i = 0;
+    search_end_index = number_of_dictionary_entries;
+
+    while (i != search_end_index)
+    {
+      search_mid_index
+        = (i + search_end_index) / 2;
+
+      TRACE_LOG("binchop-search: start:%d, mid:%d, end:%d.\n",
+          i, search_mid_index, search_end_index);
+
+      dictionary_index
+        = dictionary_start + (search_mid_index * dictionary_entry_length);
+
+      for (j=0; j<word_length; j++) {
+        TRACE_LOG("%d/%x: %x %x\n",
+            j, dictionary_index-z_mem, text_buffer[j], dictionary_index[j]);
+        if (text_buffer[j] != dictionary_index[j])
+          break;
+      }
+
+      if (j == word_length)
+      {
+        word_found_at_index = search_mid_index;
         break;
+      }
 
-    if (j == word_length)
-      break;
-
-    dictionary_index += dictionary_entry_length;
+      if (text_buffer[j] < dictionary_index[j]) {
+        TRACE_LOG("Word at %d is behind search term.\n", search_mid_index);
+        search_end_index = search_mid_index;
+      }
+      else {
+        TRACE_LOG("Word at %d is before search term.\n", search_mid_index);
+        i = search_mid_index + 1;
+      }
+    }
   }
 
-  if (i != number_of_dictionary_entries)
+  if (word_found_at_index != -1)
   {
-    TRACE_LOG("Found string at index %d.\n", i);
+    TRACE_LOG("Found string at index %d.\n", word_found_at_index);
+    /*
     TRACE_LOG("Storing %lx at %lx.\n",
-        (unsigned long int)(dictionary_index - z_mem),
+        (unsigned long int)
+        (dictionary_start[word_found_at_index] * dictionary_entry_length),
         (unsigned long int)(*parse_position - z_mem));
-    store_word(*parse_position, (uint16_t)(dictionary_index - z_mem));
+    */
+    store_word(*parse_position, (uint16_t)((dictionary_start
+          + (word_found_at_index * dictionary_entry_length)) - z_mem));
     (*parse_position) += 2;
   }
   else
@@ -546,7 +637,7 @@ static uint8_t locate_dictionary_entry(
     (*parse_position)++;
   }
 
-  if (i != number_of_dictionary_entries)
+  if (word_found_at_index != -1)
   {
     if (word_position == 1)
       first_word_found = 1;
@@ -565,18 +656,19 @@ static void tokenise(
     uint8_t *dictionary,
     bool dont_write_unrecognized_words_to_parse_buffer)
 {
-  // One normal input byte will result in a maximum of three Z-Chars: one
-  // for the '6' header chars, then 2 Z-Chars containing the value. Thus
-  // we'll need ((n*3)/3)*2 = n*2 bytes of input length: input_length * 3
-  // gives the maximum possible number of Z-Chars. One word will take up
-  // 3 Z-Chars, so we'll need a maximum of (n*3)/3 words, meaning n*2 bytes.
-  uint8_t tokenize_buffer_length = input_length * 2;
+ uint16_t dictionary_word_length = ver >= 4 ? 9 : 6;
+  uint8_t tokenize_buffer_length = ver >= 4 ? 6 : 4;
   uint8_t tokenize_buffer[tokenize_buffer_length];
 
   // The number and values of non-space-seperating input codes:
   uint8_t number_of_input_codes;
   uint8_t *input_codes;
 
+  uint8_t dictionary_entry_length;
+  int16_t number_of_dictionary_entries;
+  bool dictionary_is_unsorted;
+  uint8_t *dictionary_start;
+ 
   uint8_t current_char;
   uint8_t space_found = 0;
   uint8_t non_space_seperator_found = 0;
@@ -586,10 +678,10 @@ static void tokenise(
   uint8_t current_word_start = z_text_buffer_offset;
   uint8_t *parse_buffer_index = z_parse_buffer + 2;
   uint8_t i;
-  //int last_token_start_index;
 
-  TRACE_LOG("Parse buffer at %lx.\n",
-    (unsigned long int)(z_parse_buffer - z_mem));
+  TRACE_LOG("Parse buffer at %lx, %d bytes for tokenize_buffer at %x.\n",
+    (unsigned long int)(z_parse_buffer - z_mem), tokenize_buffer_length,
+    &tokenize_buffer);
 
   maximum_words = *z_parse_buffer;
   parse_buffer_index = z_parse_buffer + 2;
@@ -621,13 +713,9 @@ static void tokenise(
   TRACE_LOG("Number entries: %d.\n", number_of_dictionary_entries);
 #endif /* ENABLE_TRACING */
 
-  // tokenize_buffer does not have to be defined, is output buffer.
-  /*@-compdef@*/
   zchar_storage_start(tokenize_buffer, tokenize_buffer_length);
-  /*@-compdef@*/
 
   end_of_line_found = 0;
-  //last_token_start_index = 0;
   while ((number_of_words_found < maximum_words) && (end_of_line_found == 0))
   {
     TRACE_LOG("Looking at %c/%d.\n", z_text_buffer[z_text_buffer_offset],
@@ -637,18 +725,18 @@ static void tokenise(
 
     current_char = z_text_buffer[z_text_buffer_offset];
 
+    // First, determine char, if we're at the end of the input line, whether
+    // char it's a space or if char as a non-space separator.
+
     if (
         (current_char == 0)
         ||
         ( (ver >= 5) && (z_text_buffer_offset -2 == input_length) )
        )
       end_of_line_found = 1;
-
     else if (current_char == 32)
       space_found = 1;
-
     else
-    {
       for (i=0; i<number_of_input_codes; i++)
         if (current_char == input_codes[i])
         {
@@ -656,11 +744,7 @@ static void tokenise(
           break;
         }
 
-      /*
-      if ((non_space_seperator_found == 0) && (last_token_start_index == -1))
-        last_token_start_index = z_text_buffer_offset;
-        */
-    }
+    // Second, evaluate this input.
 
     if (space_found == 1)
     {
@@ -672,12 +756,15 @@ static void tokenise(
         zchar_storage_finish();
         number_of_words_found++;
 
-        (void)locate_dictionary_entry(
+        locate_dictionary_entry(
             tokenize_buffer,
             &parse_buffer_index,
             current_word_start,
             z_text_buffer_offset - current_word_start,
-            //z_text_buffer_offset - last_token_start_index - 1,
+            dictionary_start,
+            number_of_dictionary_entries,
+            dictionary_entry_length,
+            dictionary_is_unsorted,
             dont_write_unrecognized_words_to_parse_buffer);
 
         zchar_storage_start(tokenize_buffer, tokenize_buffer_length);
@@ -685,7 +772,6 @@ static void tokenise(
       space_found = 0;
       current_word_start = z_text_buffer_offset + 1;
     }
-
     else if (non_space_seperator_found == 1)
     {
       // Non-space-seperator are a bit more work, since they also count
@@ -698,65 +784,81 @@ static void tokenise(
         if (++number_of_words_found == maximum_words)
           break;
 
-        (void)locate_dictionary_entry(tokenize_buffer,
+        locate_dictionary_entry(tokenize_buffer,
             &parse_buffer_index,
             current_word_start,
             z_text_buffer_offset - current_word_start,
-            //z_text_buffer_offset - last_token_start_index - 1,
+            dictionary_start,
+            number_of_dictionary_entries,
+            dictionary_entry_length,
+            dictionary_is_unsorted,
             dont_write_unrecognized_words_to_parse_buffer);
 
         zchar_storage_start(tokenize_buffer, tokenize_buffer_length);
       }
 
+      // No length check for dictionary_word_length required, since we
+      // either just started a new zchar_storage above, or otherwise
+      // zchar_storage_symbols_stored == 0 due to if above.
       store_ZSCII_as_zchar(current_char);
       zchar_storage_finish();
       number_of_words_found++;
 
-      (void)locate_dictionary_entry(
+      locate_dictionary_entry(
           tokenize_buffer,
           &parse_buffer_index,
           z_text_buffer_offset,
           1,
+          dictionary_start,
+          number_of_dictionary_entries,
+          dictionary_entry_length,
+          dictionary_is_unsorted,
           dont_write_unrecognized_words_to_parse_buffer);
       zchar_storage_start(tokenize_buffer, tokenize_buffer_length);
 
       non_space_seperator_found = 0;
       current_word_start = z_text_buffer_offset + 1;
     }
-
     else if (end_of_line_found == 1)
     {
       zchar_storage_finish();
       if (zchar_storage_symbols_stored != 0)
       {
-        /*
-        TRACE_LOG("%x / %x / %d / %d / %d\n",
+        TRACE_LOG("%x / %x / %d / %d\n",
             tokenize_buffer,
             parse_buffer_index,
             current_word_start,
-            z_text_buffer_offset,
-            last_token_start_index);
-        */
+            z_text_buffer_offset);
         number_of_words_found++;
 
-        (void)locate_dictionary_entry(
+        locate_dictionary_entry(
             tokenize_buffer,
             &parse_buffer_index,
             current_word_start,
-            //z_text_buffer_offset - last_token_start_index - 1);
-            // line above from "kill troll ###" bug.
-            //z_text_buffer_offset - current_word_start);
             z_text_buffer_offset - current_word_start,
+            dictionary_start,
+            number_of_dictionary_entries,
+            dictionary_entry_length,
+            dictionary_is_unsorted,
             dont_write_unrecognized_words_to_parse_buffer);
       }
     }
-
     else
-      store_ZSCII_as_zchar(current_char);
+    {
+      TRACE_LOG("Symbols stored: %d.\n", zchar_storage_symbols_stored);
+      if (zchar_storage_symbols_stored < dictionary_word_length)
+        store_ZSCII_as_zchar(current_char);
+      else
+      {
+        TRACE_LOG("Maximum word length %d reached.\n", dictionary_word_length);
+      }
+    }
 
     z_text_buffer_offset++;
   }
 
+  zchar_storage_clear();
+
   TRACE_LOG("Found %d word(s), storing %d to %lx.\n", number_of_words_found,
       number_of_words_found,
       (unsigned long int)(z_parse_buffer + 1 - z_mem));
@@ -1128,9 +1230,7 @@ static void tokenise(
 
 void opcode_tokenise(void)
 {
-  /*@-nullderef@*/
   uint8_t *dictionary_table = active_z_story->dictionary_table;
-  /*@-nullderef@*/
 
   bool dont_write_unrecognized_words_to_parse_buffer = false;
   int offset = (ver >= 5 ? 2 : 1);
@@ -1341,7 +1441,6 @@ void opcode_new_line(void)
   (void)streams_latin1_output("\n");
 }
 
-
 #ifndef DISABLE_PREFIX_COMMANDS
 
 static bool process_interpreter_command()
@@ -1520,8 +1619,7 @@ static bool process_interpreter_command()
       (void)i18n_translate(
           libfizmo_module_name,
           i18n_libfizmo_P0D_BYTES_USED_BY_BLOCK_BUFFER,
-          sizeof(struct blockbuf_char) * upper_window_buffer->width
-          * upper_window_buffer->height);
+          count_allocated_blockbuf_memory(upper_window_buffer));
       (void)streams_latin1_output("\n");
     }
 #endif
@@ -1646,20 +1744,22 @@ static bool process_interpreter_command()
 
 int save_and_quit_if_required(bool force_save)
 {
-  uint8_t *filename;
+  char *filename;
   char *save_and_quit_file;
+  char *autosave_filename;
   uint8_t *pc_buf;
-  int length;
+  
+  if (active_interface->do_autosave) {
+    return active_interface->do_autosave();
+  }
 
-  save_and_quit_file
-    = get_configuration_value("save-and-quit-file-before-read");
+  autosave_filename
+    = get_configuration_value("autosave-filename");
 
-  TRACE_LOG("save_and_quit_file: %s.\n", save_and_quit_file);
+  TRACE_LOG("autosave_filename: %s.\n", autosave_filename);
 
   if (
-      (save_and_quit_file != NULL)
-      &&
-      (strcmp(save_and_quit_file, "true") == 0)
+      (autosave_filename != NULL)
       &&
       (
        (zpu_step_number != 1)
@@ -1674,25 +1774,31 @@ int save_and_quit_if_required(bool force_save)
     TRACE_LOG("current_instruction_location: %lx\n",
       (unsigned long int)(current_instruction_location - z_mem));
 
-    length = strlen(save_and_quit_file) + 1;
-
-    filename = (uint8_t*)fizmo_malloc(length);
-
-    length--;
-    memcpy(filename + 1, save_and_quit_file, length);
-    *filename = length;
+    filename = fizmo_strdup(autosave_filename);
 
     save_game(
       0,
       (uint16_t)(active_z_story->dynamic_memory_end - z_mem + 1),
-      save_and_quit_file,
+      filename,
       true,
       false,
       NULL);
 
     pc = pc_buf;
-    terminate_interpreter = INTERPRETER_QUIT_SAVE_BEFORE_READ;
-    return 1;
+
+    save_and_quit_file
+      = get_configuration_value("save-and-quit-file-before-read");
+    if (
+        (save_and_quit_file != NULL)
+        &&
+        (strcmp(save_and_quit_file, "true") == 0)) 
+    {
+      terminate_interpreter = INTERPRETER_QUIT_SAVE_BEFORE_READ;
+      return 1;
+    }
+    else {
+      return 0;
+    }
   }
   else
     return 0;
@@ -1707,21 +1813,42 @@ int read_command_from_file(zscii *input_buffer, int input_buffer_size,
   long filepos;
   int milliseconds;
   int res;
+  int return_code;
 
   if (input_stream_1 == NULL)
   {
-    if (input_stream_1_was_already_active == false)
-      ask_for_input_stream_filename();
-    TRACE_LOG("Trying to open \"%s\"\n", input_stream_1_filename);
-    if ((input_stream_1 = fsi->openfile(input_stream_1_filename,
-            FILETYPE_INPUTRECORD, FILEACCESS_READ)) == NULL)
-    {
-      TRACE_LOG("Could not open input file for stream 1.\n");
-      input_stream_1_active = false;
-      input_stream_1_filename_size = 0;
-      free(input_stream_1_filename);
-      return -1;
-    }
+      return_code = active_interface->prompt_for_filename(
+          "transcript",
+          &input_stream_1,
+          NULL,
+          FILETYPE_TRANSCRIPT,
+          FILEACCESS_READ);
+
+      if (return_code == -3)
+      {
+        // If return_code is == -3, this function is not implemented in the
+        // current screen interface.
+
+        if (input_stream_1_was_already_active == false)
+        {
+          ask_for_input_stream_filename();
+        }
+
+        TRACE_LOG("Trying to open \"%s\"\n", input_stream_1_filename);
+        if ((input_stream_1 = fsi->openfile(input_stream_1_filename,
+                FILETYPE_INPUTRECORD, FILEACCESS_READ)) == NULL)
+        {
+          TRACE_LOG("Could not open input file for stream 1.\n");
+          input_stream_1_active = false;
+          input_stream_1_filename_size = 0;
+          free(input_stream_1_filename);
+          return -1;
+        }
+      }
+      else if (return_code < 0)
+      {
+        return -1;
+      }
   }
 
   filepos = fsi->getfilepos(input_stream_1);
@@ -1742,7 +1869,7 @@ int read_command_from_file(zscii *input_buffer, int input_buffer_size,
 
   // FIMXE: Input conversion from latin1(?) to zscii-input.
   while (
-      ((c = fsi->getchar(input_stream_1)) != EOF)
+      ((c = fsi->readchar(input_stream_1)) != EOF)
       &&
       (c != '\n')
       &&
@@ -1760,7 +1887,7 @@ int read_command_from_file(zscii *input_buffer, int input_buffer_size,
   {
     // Seek next '\n'.
     while (c != '\n')
-      if ((c = fsi->getchar(input_stream_1)) == EOF)
+      if ((c = fsi->readchar(input_stream_1)) == EOF)
         break;
   }
 
@@ -1776,7 +1903,7 @@ int read_command_from_file(zscii *input_buffer, int input_buffer_size,
   }
 
   // Check if there's something behind the '\n'.
-  c = fsi->getchar(input_stream_1);
+  c = fsi->readchar(input_stream_1);
   if (c == EOF)
   {
     input_stream_1_active = false;
@@ -1784,7 +1911,7 @@ int read_command_from_file(zscii *input_buffer, int input_buffer_size,
     input_stream_1 = NULL;
   }
   else
-    fsi->ungetchar(c, input_stream_1);
+    fsi->unreadchar(c, input_stream_1);
 
   return input_length;
 }
@@ -1798,9 +1925,13 @@ void opcode_read(void)
   uint16_t parsebuffer_offset = op[1];
   uint16_t tenth_seconds = op[2];
   uint16_t timed_routine_offset = op[3];
+  zscii nof_preloaded_chars;
   bool interpreter_command_found;
   int i;
-  z_ucs *current_line;
+  // TODO: Implement permanent buffer for current_line.
+#ifndef DISABLE_OUTPUT_HISTORY
+  z_ucs *current_line = NULL;
+#endif /* DISABLE_OUTPUT_HISTORY */
   bool stream_1_active_buf;
   int offset;
   size_t bytes_required;
@@ -1820,7 +1951,6 @@ void opcode_read(void)
 
   TRACE_LOG("Reading input (%x, %x).\n", op[0], parsebuffer_offset);
 
-  // FIXME: Implement permanent buffer for current_line.
 #ifndef DISABLE_OUTPUT_HISTORY
   current_line = get_current_line(outputhistory[active_window_number]);
   if (current_line != NULL)
@@ -1829,8 +1959,6 @@ void opcode_read(void)
     TRACE_LOG_Z_UCS(current_line);
     TRACE_LOG("'.\n");
   }
-#else
-  current_line = NULL;
 #endif /* DISABLE_OUTPUT_HISTORY */
 
   if (ver >= 5)
@@ -1905,7 +2033,8 @@ void opcode_read(void)
 #endif // ENABLE_TRACING
 
       // Preloaded input is only availiable in version 5.
-      if (z_text_buffer[1] >  0)
+      nof_preloaded_chars = z_text_buffer[1];
+      if (nof_preloaded_chars > 0)
       {
         if (active_interface->is_preloaded_input_available() == false)
           i18n_translate_and_exit(
@@ -1914,9 +2043,10 @@ void opcode_read(void)
               -1,
               active_interface->get_interface_name());
 
-        TRACE_LOG("Removing %d chars from history.\n", z_text_buffer[1]);
+        TRACE_LOG("Removing %d chars from history.\n", nof_preloaded_chars);
+        stream_2_remove_chars(nof_preloaded_chars);
 #ifndef DISABLE_OUTPUT_HISTORY
-        remove_chars_from_history(outputhistory[0], z_text_buffer[1]);
+        remove_chars_from_history(outputhistory[0], nof_preloaded_chars);
 #endif // DISABLE_OUTPUT_HISTORY
       }
 
@@ -1945,6 +2075,7 @@ void opcode_read(void)
               TRACE_LOG(
                   "Invoking verification routine, 1/10s counter: %d(%d).\n",
                   tenth_seconds_to_delay , tenth_seconds);
+#ifndef DISABLE_OUTPUT_HISTORY
               if ((stream_output_has_occured == true) && (current_line != NULL))
               {
                 TRACE_LOG("Restoring original prompt: '");
@@ -1953,6 +2084,7 @@ void opcode_read(void)
                 (void)streams_z_ucs_output(current_line);
                 stream_output_has_occured = false;
               }
+#endif // DISABLE_OUTPUT_HISTORY
               if ((timed_routine_retval=interpret_from_call(
                       get_packed_routinecall_address(timed_routine_offset)))!=0)
               {
@@ -1986,7 +2118,13 @@ void opcode_read(void)
         }
 
         if (terminate_interpreter != INTERPRETER_QUIT_NONE)
+        {
+#ifndef DISABLE_OUTPUT_HISTORY
+          if (current_line != NULL)
+            free(current_line);
+#endif // DISABLE_OUTPUT_HISTORY
           return;
+        }
 
         // input_length always defined, since "i18n_translate_and_exit" does
         // not return.
@@ -2107,7 +2245,7 @@ void opcode_read(void)
         tokenise(
             z_text_buffer,
             (uint8_t)(ver >= 5 ? 2 : 1),
-            /*@-usedef@*/ (uint16_t)input_length /*@+usedef@*/,
+            (uint16_t)input_length,
             z_mem + parsebuffer_offset,
             active_z_story->dictionary_table,
             false);
@@ -2120,6 +2258,7 @@ void opcode_read(void)
 
         if (bool_equal(interpreter_command_found, true))
         {
+#ifndef DISABLE_OUTPUT_HISTORY
           if (current_line != NULL)
           {
             TRACE_LOG("Restoring original prompt: '");
@@ -2128,6 +2267,7 @@ void opcode_read(void)
             (void)streams_z_ucs_output(current_line);
           }
           else
+#endif // DISABLE_OUTPUT_HISTORY
             (void)streams_latin1_output(">");
 
           z_text_buffer[1] = 0;
@@ -2154,6 +2294,7 @@ void opcode_read(void)
   // interpreter, not the story, we'll continue with the input.
   while (interpreter_command_found != false);
 
+#ifndef DISABLE_OUTPUT_HISTORY
   if (current_line != NULL)
   {
     TRACE_LOG("Free current line (%p): '", current_line);
@@ -2161,6 +2302,7 @@ void opcode_read(void)
     TRACE_LOG("'.\n");
     free(current_line);
   }
+#endif // DISABLE_OUTPUT_HISTORY
 
   number_of_commands++;
 }
@@ -2413,6 +2555,7 @@ void opcode_encode_text(void)
   }
 
   zchar_storage_finish();
+  zchar_storage_clear();
 }
 
 
