@@ -16,7 +16,7 @@
  *    documentation and/or other materials provided with the distribution.
  * 3. The name of the author may not be used to endorse or promote products
  *    derived from this software without specific prior written permission.
- * 
+ *
  * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
  * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
@@ -39,6 +39,7 @@
 #include <strings.h>
 
 #include "glk.h"
+#include "gi_dispa.h"
 #include "glkstart.h" /* This comes with the Glk library. */
 
 #include "glk_interface.h"
@@ -54,26 +55,73 @@
 #include <tools/tracelog.h>
 
 static char* interface_name = "glk-screen";
-static char* interface_version = "0.1.2";
+static char* interface_version = "0.1.6";
+
+static z_file *(*game_open_interface)(z_file *) = NULL;
+static z_file *story_stream = NULL;
 
 static winid_t mainwin = NULL;
 static winid_t statusline = NULL;
 static winid_t statuswin = NULL;
 static bool instatuswin = false;
+static int statuscurheight = 0; /* what the VM thinks the height is */
+static int statusmaxheight = 0; /* height including possible quote box */
+static int statusseenheight = 0; /* last height the user saw */
+static int screenestwidth = 0; /* width of screen in status-line characters */
+static int screenestheight = 0; /* height of screen in status-line characters */
 static int inputbuffer_size = 0;
 static glui32 *inputbuffer = NULL;
 
-static void glkint_get_screen_size(glui32 *, glui32 *);
+static void glkint_resolve_status_height(void);
+static void glkint_estimate_screen_size(void);
+static char *glkint_get_game_id(void);
 
-void glkint_open_interface()
+z_file *glkint_open_interface(z_file *(*game_open_func)(z_file *))
 {
+  /* Clear all of the static variables. We might be restarting (on iOS)
+     so we can't rely on the static initializers. */
+  mainwin = NULL;
+  statusline = NULL;
+  statuswin = NULL;
+  instatuswin = false;
+  statuscurheight = 0; 
+  statusmaxheight = 0;
+  statusseenheight = 0;
+  screenestwidth = 0;
+  screenestheight = 0;
+  /* Skip inputbuffer; that's just a malloced block and a size, so it can 
+     persist across restarts. */
+
+  /* Set up the game-ID hook. (This is ifdeffed because not all Glk
+     libraries have this call.) */
+#ifdef GI_DISPA_GAME_ID_AVAILABLE
+  gidispatch_set_game_id_hook(&glkint_get_game_id);
+#endif /* GI_DISPA_GAME_ID_AVAILABLE */
+
+  /* The awkward nature of iOS autosave-restore means that we need to
+     retain a reference to the story_stream, and possibly fix it up
+     later on. If this isn't iOS, just ignore this juggling. */
+  game_open_interface = game_open_func;
+  story_stream = game_open_interface(NULL);
+  if (!story_stream) {
+    return NULL;
+  }
+
   mainwin = glk_window_open(NULL, 0, 0, wintype_TextBuffer, 1);
   glk_set_window(mainwin);
   instatuswin = false;
 
-  glui32 width, height;
-  glkint_get_screen_size( &width, &height);
-  fizmo_new_screen_size(width, height);
+  if (1) {
+      /* First approximation of the screen size has to be based on the
+         story window. When a status window is opened, we'll refine
+         this. */
+      glui32 width, height;
+      glk_window_get_size(mainwin, &width, &height);
+      screenestwidth = width;
+      screenestheight = height;
+  }
+  
+  return story_stream;
 }
 
 char *glkint_get_interface_name()
@@ -91,34 +139,95 @@ uint8_t glkint_return_0()
 uint8_t glkint_return_1()
 { return 1; }
 
-static void glkint_get_screen_size(glui32 *width, glui32 *height)
+/* This is called after an autosave-restore (iOS only). We've just
+   pulled a new Glk library state from disk. We need to go through it
+   and set mainwin, statuswin, etc appropriately.
+
+   Note that at this point, the Glk streams opened by the interpreter
+   (the original story file, and a transcript stream if any) have been
+   closed. We'll need to update the interpreter with replacements.
+ */
+void glkint_recover_library_state()
 {
+  strid_t storystream = NULL;
+  strid_t transcriptstream = NULL;
+  z_file *transcriptzfile = NULL;
+  winid_t win;
+  strid_t str;
+  glui32 rock;
+
+  mainwin = NULL;
+  statusline = NULL;
+  statuswin = NULL;
+  instatuswin = false;
+  zfile_replace_glk_strid(story_stream, NULL);
+
+  /* Close the old transcript stream, if there was one. */
+  transcriptzfile = get_stream_2();
+  /* This is a little bit fiddly, because the underlying Glk stream is
+     already gone. We mark this by removing the pointer to it, and
+     then the z_file can be closed safely. */
+  if (transcriptzfile) {
+    zfile_replace_glk_strid(transcriptzfile, NULL);
+    restore_stream_2(NULL);
+  }
+
+  statuscurheight = 0;
+  statusmaxheight = 0;
+  statusseenheight = 0;
+
+  win = NULL;
+  while ((win=glk_window_iterate(win, &rock)) != NULL) {
+    if (rock == 1)
+      mainwin = win;
+    else if (rock == 2)
+      statuswin = win;
+    else if (rock == 3)
+      statusline = win;
+  }
+
   if (statuswin) {
-    glk_window_get_size(statuswin, width, height);
-    return;
+    glui32 truewidth, trueheight;
+    glk_window_get_size(statuswin, &truewidth, &trueheight);
+    statuscurheight = trueheight;
+    statusmaxheight = trueheight;
+    statusseenheight = trueheight;
   }
-  if (mainwin) {
-    glk_window_get_size(mainwin, width, height);
-    return;
+
+  str = NULL;
+  while ((str=glk_stream_iterate(str, &rock)) != NULL) {
+    if (rock == 1)
+      storystream = str;
+    else if (rock == 4)
+      transcriptstream = str;
   }
-  /* Fallback values, for when no windows are open at all */
-  *width = 80;
-  *height = 24;
-  return;
+
+  /* Close the old story stream which we found in the library state. (We're
+     about to open a fresh version.) */
+  if (storystream) {
+    glk_stream_close(storystream, NULL);
+  }
+
+  game_open_interface(story_stream);
+
+  /* If we found a transcript stream, pass it to the library. */
+  if (transcriptstream) {
+    transcriptzfile = zfile_from_glk_strid(transcriptstream, NULL,
+      FILETYPE_TRANSCRIPT, FILEACCESS_APPEND);
+    restore_stream_2(transcriptzfile);
+  }
+  
+  glkint_estimate_screen_size();
 }
 
 uint8_t glkint_get_screen_height()
-{ 
-  glui32 width, height;
-  glkint_get_screen_size(&width, &height);
-  return height;
+{
+  return screenestheight;
 }
 
 uint8_t glkint_get_screen_width()
-{ 
-  glui32 width, height;
-  glkint_get_screen_size(&width, &height);
-  return width;
+{
+  return screenestwidth;
 }
 
 z_colour glkint_get_default_foreground_colour()
@@ -141,7 +250,7 @@ char **glkint_get_config_option_names()
 }
 
 void glkint_link_interface_to_story(struct z_story *UNUSED(story))
-{ 
+{
   if (ver <= 3)
   {
     if (statusline) {
@@ -150,10 +259,10 @@ void glkint_link_interface_to_story(struct z_story *UNUSED(story))
     statusline = glk_window_open(
         mainwin, winmethod_Above | winmethod_Fixed, 1, wintype_TextGrid, 3);
     /*
-    glk_set_window(statusline);    
+    glk_set_window(statusline);
     glk_set_style(style_Normal | stylehint_ReverseColor);
     glk_window_clear(statusline);
-    glk_set_window(mainwin);    
+    glk_set_window(mainwin);
     */
   }
 }
@@ -167,14 +276,29 @@ void glkint_reset_interface()
   }
 
   instatuswin = false;
-  glk_set_window(mainwin);    
+  glk_set_window(mainwin);
   glk_set_style(style_Normal);
   glk_window_clear(mainwin);
+  
+  /* Leave the screen-size estimates alone. Closing the status window
+     doesn't give us new information. */
 }
 
-/* Called at @quit time, or if the interpreter hits a fatal error. */
+/* This is called from two points: abort_interpreter() with an error message,
+   and close_streams() with no error message.
+
+   For the abort case, we call glkint_fatal_error_handler(), which does a
+   glk_exit. This is necessary because abort_interpreter() is going to do
+   a libc exit(), and we need to get Glk shut down first.
+
+   However, close_streams() is different -- that's the normal interpreter
+   exit (a @quit opcode). For that case, it's better to return and do nothing,
+   so that close_streams() can flush fizmo's buffers. It will immediately
+   exit through the end of fizmo_main(), which is the end of glk_main(),
+   so we'll get a normal Glk shut down anyway.
+   */
 int glkint_close_interface(z_ucs *error_message)
-{ 
+{
   if (error_message)
     glkint_fatal_error_handler(NULL, error_message, NULL, 0, 0);
   return 0;
@@ -198,6 +322,8 @@ int16_t glkint_interface_read_line(zscii *dest, uint16_t maximum_length,
   zscii zch;
   int i;
 
+  glkint_resolve_status_height();
+
   if (!inputbuffer) {
     inputbuffer_size = maximum_length+16;
     inputbuffer = malloc(inputbuffer_size * sizeof(glui32));
@@ -232,9 +358,7 @@ int16_t glkint_interface_read_line(zscii *dest, uint16_t maximum_length,
       break;
 
     if (event.type == evtype_Arrange) {
-      glui32 width, height;
-      glkint_get_screen_size( &width, &height);
-      fizmo_new_screen_size(width, height);
+      glkint_estimate_screen_size();
       continue;
     }
 
@@ -275,7 +399,11 @@ int glkint_interface_read_char(uint16_t tenth_seconds,
   event_t event;
   int timercount = 0;
   int timed_routine_retval;
-  winid_t win = (instatuswin ? statuswin : mainwin);
+  winid_t win;
+
+  glkint_resolve_status_height();
+
+  win = (instatuswin ? statuswin : mainwin);
 
   if (win) {
     glk_request_char_event_uni(win);
@@ -292,9 +420,7 @@ int glkint_interface_read_char(uint16_t tenth_seconds,
       break;
 
     if (event.type == evtype_Arrange) {
-      glui32 width, height;
-      glkint_get_screen_size( &width, &height);
-      fizmo_new_screen_size(width, height);
+      glkint_estimate_screen_size();
       continue;
     }
 
@@ -348,7 +474,7 @@ void glkint_show_status(z_ucs *room_description,
 {
   char buf[128];
 
-  glk_set_window(statusline);    
+  glk_set_window(statusline);
   glk_window_clear(statusline);
   glk_set_style(style_Subheader);
   glk_put_string_uni(room_description);
@@ -365,11 +491,11 @@ void glkint_show_status(z_ucs *room_description,
     glk_put_string(buf);
   }
 
-  glk_set_window(mainwin);    
+  glk_set_window(mainwin);
 }
 
 void glkint_set_text_style(z_style text_style)
-{ 
+{
   if (text_style & Z_STYLE_FIXED_PITCH) {
     glk_set_style(style_Preformatted);
   }
@@ -389,7 +515,7 @@ void glkint_set_colour(z_colour UNUSED(foreground),
 { }
 
 void glkint_set_font(z_font font_type)
-{ 
+{
   if (instatuswin)
     return;
   if (font_type == Z_FONT_COURIER_FIXED_PITCH)
@@ -399,37 +525,107 @@ void glkint_set_font(z_font font_type)
 }
 
 void glkint_split_window(int16_t nof_lines)
-{ 
-  if (!nof_lines) {
-    if (statuswin) {
+{
+  int oldvmheight = statuscurheight;
+  statuscurheight = nof_lines;
+
+  /* We do not decrease the height at this time -- it can only increase.
+     This ensures that quote boxes don't vanish. */
+  if (statuscurheight > statusmaxheight)
+    statusmaxheight = statuscurheight;
+
+  /* However, if the VM thinks it's increasing the height, we must be
+     careful to clear the "newly created" space. */
+  if (statuswin && statuscurheight > oldvmheight) {
+    strid_t stream = glk_window_get_stream(statuswin);
+    int ix, jx;
+    glui32 truewidth, trueheight;
+    glk_window_get_size(statuswin, &truewidth, &trueheight);
+    for (jx=oldvmheight; jx<(int)trueheight; jx++) {
+      glk_window_move_cursor(statuswin, 0, jx);
+      for (ix=0; ix<(int)truewidth; ix++) {
+        glk_put_char_stream(stream, ' ');
+      }
+    }
+    glk_window_move_cursor(statuswin, 0, 0);
+  }
+
+  if (!statuswin) {
+    statuswin = glk_window_open(mainwin, winmethod_Above | winmethod_Fixed, statusmaxheight, wintype_TextGrid, 2);
+  }
+  else {
+    glk_window_set_arrangement(glk_window_get_parent(statuswin), winmethod_Above | winmethod_Fixed, statusmaxheight, statuswin);
+  }
+  
+  glkint_estimate_screen_size();
+}
+
+/* If the status height is too large because of last turn's quote box,
+   shrink it down now. */
+static void glkint_resolve_status_height()
+{
+  if (statusseenheight == statusmaxheight)
+    statusmaxheight = statuscurheight;
+
+  if (statuswin) {
+    if (statusmaxheight == 0) {
       glk_window_close(statuswin, NULL);
       statuswin = NULL;
     }
-  }
-  else {
-    if (!statuswin) {
-      statuswin = glk_window_open(mainwin, winmethod_Above | winmethod_Fixed, nof_lines, wintype_TextGrid, 2);
-    }
     else {
-      glk_window_set_arrangement(glk_window_get_parent(statuswin), winmethod_Above | winmethod_Fixed, nof_lines, statuswin);
+      glk_window_set_arrangement(glk_window_get_parent(statuswin), winmethod_Above | winmethod_Fixed, statusmaxheight, statuswin);
     }
   }
+
+  statusseenheight = statusmaxheight;
+  statusmaxheight = statuscurheight;
+
+  glkint_estimate_screen_size();
+}
+
+/* Construct an estimate of the current screen size, and pass it to the VM.
+ 
+   The estimate uses the width (in status chars) of the status window --
+   that's what's important for 99% of games. For the height, we add
+   the number of lines in the status and story windows; this won't be
+   perfect (because the fonts are different) but it's close enough.
+ */
+static void glkint_estimate_screen_size()
+{
+    glui32 width, height;
+    
+    if (!statuswin) {
+        /* Leave the screen-size estimates alone. Closing the status
+           window doesn't give us new information. It's better to keep
+           using an old exact measurement of the status window than to
+           approximate it in the story window. */
+        return;
+    }
+    
+    glk_window_get_size(statuswin, &width, &height);
+    screenestwidth = width;
+    screenestheight = height;
+    
+    glk_window_get_size(mainwin, &width, &height);
+    screenestheight += height;
+    
+    fizmo_new_screen_size(screenestwidth, screenestheight);
 }
 
 /* 1 is the status window; 0 is the story window. */
 void glkint_set_window(int16_t window_number)
-{ 
+{
   if (!window_number) {
     glk_set_window(mainwin);
     instatuswin = false;
   }
   else {
-    if (statuswin) 
+    if (statuswin)
       glk_set_window(statuswin);
     else
       glk_set_window(NULL);
     instatuswin = true;
-  }    
+  }
 }
 
 void glkint_erase_window(int16_t window_number)
@@ -445,7 +641,7 @@ void glkint_erase_window(int16_t window_number)
 
 void glkint_set_cursor(int16_t line, int16_t column,
     int16_t window)
-{ 
+{
   if (window && statuswin)
     glk_window_move_cursor(statuswin, column-1, line-1);
 }
@@ -475,6 +671,112 @@ void glkint_output_interface_info()
 void glkint_game_was_restored_and_history_modified()
 { }
 
+int glkint_prompt_for_filename(char *UNUSED(filename_suggestion),
+    z_file **result, char *UNUSED(directory), int filetype, int fileaccess)
+{
+  frefid_t fileref = NULL;
+  strid_t str = NULL;
+  glui32 usage, fmode, strrock;
+
+  if (filetype == FILETYPE_SAVEGAME)
+  {
+    usage = fileusage_SavedGame | fileusage_BinaryMode;
+    strrock = 3;
+  }
+  else if (filetype == FILETYPE_TRANSCRIPT)
+  {
+    usage = fileusage_Transcript | fileusage_TextMode;
+    strrock = 4;
+  }
+  else if (filetype == FILETYPE_INPUTRECORD)
+  {
+    usage = fileusage_InputRecord | fileusage_TextMode;
+    strrock = 5;
+  }
+  else if (filetype == FILETYPE_DATA)
+  {
+    usage = fileusage_Data | fileusage_BinaryMode;
+    strrock = 6;
+  }
+  else if (filetype == FILETYPE_TEXT)
+  {
+    usage = fileusage_Data | fileusage_TextMode;
+    strrock = 7;
+  }
+  else
+    return -1;
+
+  if (fileaccess == FILEACCESS_READ)
+    fmode = filemode_Read;
+  else if (fileaccess == FILEACCESS_WRITE)
+    fmode = filemode_Write;
+  else if (fileaccess == FILEACCESS_APPEND)
+    fmode = filemode_WriteAppend;
+  else
+    return -1;
+
+  fileref = glk_fileref_create_by_prompt(usage, fmode, 0);
+
+  if (!fileref)
+    return -1;
+
+  str = glk_stream_open_file(fileref, fmode, strrock);
+  /* Dispose of the fileref, whether the stream opened successfully
+   * or not. */
+  glk_fileref_destroy(fileref);
+
+  if (!str)
+    return -1;
+
+  *result = zfile_from_glk_strid(
+      str, NULL, // cannot fetch filename from fileref?
+      filetype, fileaccess);
+
+  return 0;
+}
+
+static char *glkint_get_game_id()
+{
+  /* This buffer gets rewritten on every call, but that's okay -- the caller
+     is supposed to copy out the result. */
+  static char buf[32];
+  char ch;
+  int ix;
+  int jx;
+
+  if (!z_mem)
+    return NULL;
+
+  /* This is computed in a way similar to the game-identifier chunk 
+     for the save file. */
+  
+  jx = 0;
+
+  /* Release number: */
+  for (ix=2; ix<4; ix++) {
+    ch = z_mem[ix];
+    buf[jx++] = (((ch >> 4) & 0x0F) + 'A');
+    buf[jx++] = ((ch & 0x0F) + 'A');
+  }
+
+  /* Serial number: */
+  for (ix=0x12; ix<0x18; ix++) {
+    ch = z_mem[ix];
+    buf[jx++] = (((ch >> 4) & 0x0F) + 'A');
+    buf[jx++] = ((ch & 0x0F) + 'A');
+  }
+
+  /* Checksum: */
+  for (ix=0x1C; ix<0x1E; ix++) {
+    ch = z_mem[ix];
+    buf[jx++] = (((ch >> 4) & 0x0F) + 'A');
+    buf[jx++] = ((ch & 0x0F) + 'A');
+  }
+
+  buf[jx++] = '\0';
+  return buf;
+}
+
 struct z_screen_interface glkint_screen_interface =
 {
   &glkint_get_interface_name,
@@ -523,7 +825,10 @@ struct z_screen_interface glkint_screen_interface =
   &glkint_erase_line_pixels,
   &glkint_output_interface_info,
   &glkint_return_false, /* input_must_be_repeated_by_story */
-  &glkint_game_was_restored_and_history_modified
+  &glkint_game_was_restored_and_history_modified,
+  &glkint_prompt_for_filename,
+  NULL, /* do_autosave */
+  NULL /* restore_autosave */
 };
 
 #endif // glk_screen_if_c_INCLUDED
